#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from enum import Flag, auto
from typing import Any, List, Union

import merlin
from merlin.dag.selector import ColumnSelector
from merlin.schema import ColumnSchema, Schema


class Supports(Flag):
    """Indicates what type of data representation this operator supports for transformations"""

    # cudf dataframe
    CPU_DATAFRAME = auto()
    # pandas dataframe
    GPU_DATAFRAME = auto()
    # dict of column name to numpy array
    CPU_DICT_ARRAY = auto()
    # dict of column name to cupy array
    GPU_DICT_ARRAY = auto()


class BaseOperator:
    """
    Base class for all operator classes.
    """

    def compute_selector(
        self,
        input_schema: Schema,
        selector: ColumnSelector,
        parents_selector: ColumnSelector,
        dependencies_selector: ColumnSelector,
    ) -> ColumnSelector:
        self._validate_matching_cols(input_schema, selector, self.compute_selector.__name__)

        return selector

    def compute_input_schema(
        self,
        root_schema: Schema,
        parents_schema: Schema,
        deps_schema: Schema,
        selector: ColumnSelector,
    ) -> Schema:
        """Given the schemas coming from upstream sources and a column selector for the
        input columns, returns a set of schemas for the input columns this operator will use
        Parameters
        -----------
        root_schema: Schema
            Base schema of the dataset before running any operators.
        parents_schema: Schema
            The combined schemas of the upstream parents feeding into this operator
        deps_schema: Schema
            The combined schemas of the upstream dependencies feeding into this operator
        col_selector: ColumnSelector
            The column selector to apply to the input schema
        Returns
        -------
        Schema
            The schemas of the columns used by this operator
        """
        self._validate_matching_cols(
            parents_schema + deps_schema, selector, self.compute_input_schema.__name__
        )

        return parents_schema + deps_schema

    def compute_output_schema(
        self,
        input_schema: Schema,
        col_selector: ColumnSelector,
        prev_output_schema: Schema = None,
    ) -> Schema:
        """Given a set of schemas and a column selector for the input columns,
        returns a set of schemas for the transformed columns this operator will produce
        Parameters
        -----------
        input_schema: Schema
            The schemas of the columns to apply this operator to
        col_selector: ColumnSelector
            The column selector to apply to the input schema
        Returns
        -------
        Schema
            The schemas of the columns produced by this operator
        """
        if not col_selector:
            col_selector = ColumnSelector(input_schema.column_names)

        if col_selector.tags:
            tags_col_selector = ColumnSelector(tags=col_selector.tags)
            filtered_schema = input_schema.apply(tags_col_selector)
            col_selector += ColumnSelector(filtered_schema.column_names)

            # zero tags because already filtered
            col_selector._tags = []

        self._validate_matching_cols(
            input_schema, col_selector, self.compute_output_schema.__name__
        )

        output_schema = Schema()
        for output_col_name, input_col_names in self.column_mapping(col_selector).items():
            col_schema = ColumnSchema(output_col_name)
            col_schema = self._compute_dtype(col_schema, input_schema[input_col_names])
            col_schema = self._compute_tags(col_schema, input_schema[input_col_names])
            col_schema = self._compute_properties(col_schema, input_schema[input_col_names])
            output_schema += Schema([col_schema])

        if self.dynamic_dtypes and prev_output_schema:
            for col_name, col_schema in output_schema.column_schemas.items():
                dtype = prev_output_schema[col_name].dtype
                output_schema.column_schemas[col_name] = col_schema.with_dtype(dtype)

        return output_schema

    def column_mapping(self, col_selector):
        column_mapping = {}
        for col_name in col_selector.names:
            column_mapping[col_name] = [col_name]
        return column_mapping

    def compute_column_schema(self, col_name, input_schema):
        methods = [self._compute_dtype, self._compute_tags, self._compute_properties]
        return self._compute_column_schema(col_name, input_schema, methods=methods)

    def _compute_column_schema(self, col_name, input_schema, methods=None):
        col_schema = ColumnSchema(col_name)

        for method in methods:
            col_schema = method(col_schema, input_schema)

        return col_schema

    def _compute_dtype(self, col_schema, input_schema):
        dtype = col_schema.dtype
        is_list = col_schema.is_list
        is_ragged = col_schema.is_ragged

        if input_schema.column_schemas:
            source_col_name = input_schema.column_names[0]
            dtype = input_schema[source_col_name].dtype
            is_list = input_schema[source_col_name].is_list
            is_ragged = input_schema[source_col_name].is_ragged

        if self.output_dtype is not None:
            dtype = self.output_dtype
            is_list = any(cs.is_list for _, cs in input_schema.column_schemas.items())
            is_ragged = any(cs.is_ragged for _, cs in input_schema.column_schemas.items())

        return col_schema.with_dtype(dtype, is_list=is_list, is_ragged=is_ragged)

    @property
    def dynamic_dtypes(self):
        return False

    def _compute_tags(self, col_schema, input_schema):
        tags = []
        if input_schema.column_schemas:
            source_col_name = input_schema.column_names[0]
            tags = input_schema[source_col_name].tags

        # Override empty tags with tags from the input schema
        # Override input schema tags with the output tags of this operator
        return col_schema.with_tags(tags).with_tags(self.output_tags)

    def _compute_properties(self, col_schema, input_schema):
        properties = {}

        if input_schema.column_schemas:
            source_col_name = input_schema.column_names[0]
            properties.update(input_schema.column_schemas[source_col_name].properties)

        properties.update(self.output_properties)

        return col_schema.with_properties(properties)

    def _validate_matching_cols(self, schema, selector, method_name):
        selector = selector or ColumnSelector()
        missing_cols = [name for name in selector.names if name not in schema.column_names]
        if missing_cols:
            raise ValueError(
                f"Missing columns {missing_cols} found in operator"
                f"{self.__class__.__name__} during {method_name}."
            )

    # TODO: Update instructions for how to define custom
    # operators to reflect constructing the column mapping
    # (They should no longer override this method)
    def output_column_names(self, col_selector: ColumnSelector) -> ColumnSelector:
        """Given a set of columns names returns the names of the transformed columns this
        operator will produce
        Parameters
        -----------
        columns: list of str, or list of list of str
            The columns to apply this operator to
        Returns
        -------
        list of str, or list of list of str
            The names of columns produced by this operator
        """
        return ColumnSelector(list(self.column_mapping(col_selector).keys()))

    @property
    def dependencies(self) -> List[Union[str, Any]]:
        """Defines an optional list of column dependencies for this operator. This lets you consume columns
        that aren't part of the main transformation workflow.
        Returns
        -------
        str, list of str or ColumnSelector, optional
            Extra dependencies of this operator. Defaults to None
        """
        return []

    def __rrshift__(self, other):
        return ColumnSelector(other) >> self

    @property
    def output_dtype(self):
        return None

    @property
    def output_tags(self):
        return []

    @property
    def output_properties(self):
        return {}

    @property
    def label(self) -> str:
        return self.__class__.__name__

    def create_node(self, selector):
        return merlin.dag.Node(selector)

    @property
    def supports(self) -> Supports:
        """Returns what kind of data representation this operator supports"""
        return Supports.CPU_DATAFRAME | Supports.GPU_DATAFRAME

    def _get_columns(self, df, selector):
        if isinstance(df, dict):
            return {col_name: df[col_name] for col_name in selector.names}
        else:
            return df[selector.names]



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools

import dask.dataframe as dd

try:
    import dask_cudf
except ImportError:
    dask_cudf = None
import numpy as np
from dask.bytes import read_bytes
from dask.utils import parse_bytes
from fsspec.core import get_fs_token_paths
from fsspec.utils import infer_compression

from .dataset_engine import DatasetEngine


class CSVDatasetEngine(DatasetEngine):
    """CSVDatasetEngine

    Thin wrapper around dask_cudf.read_csv.
    """

    def __init__(self, paths, part_size, storage_options=None, cpu=False, **kwargs):
        super().__init__(paths, part_size, cpu=cpu, storage_options=storage_options)
        self._meta = {}
        self.csv_kwargs = kwargs
        self.csv_kwargs["storage_options"] = storage_options

        # CSV reader needs a list of files
        # (Assume flat directory structure if this is a dir)
        if len(self.paths) == 1 and self.fs.isdir(self.paths[0]):
            self.paths = self.fs.glob(self.fs.sep.join([self.paths[0], "*"]))

    def to_ddf(self, columns=None, cpu=None):

        # Check if we are using cpu
        cpu = self.cpu if cpu is None else cpu
        if cpu:
            ddf = dd.read_csv(self.paths, blocksize=self.part_size, **self.csv_kwargs)
        else:
            ddf = dask_cudf.read_csv(self.paths, chunksize=self.part_size, **self.csv_kwargs)
        if columns:
            ddf = ddf[columns]
        return ddf

    @property
    @functools.lru_cache(1)
    def _file_partition_map(self):
        ind = 0
        _pp_map = {}
        for path, blocks in zip(
            *_byte_block_counts(
                self.paths,
                self.part_size,
                **self.csv_kwargs,
            )
        ):
            _pp_map[path.split(self.fs.sep)[-1]] = np.arange(ind, ind + blocks)
            ind += blocks
        return _pp_map

    def to_cpu(self):
        self.cpu = True

    def to_gpu(self):
        self.cpu = False


def _byte_block_counts(
    urlpath,
    blocksize,
    lineterminator=None,
    compression="infer",
    storage_options=None,
    **kwargs,
):
    """Return a list of paths and block counts.

    Logic copied from dask.bytes.read_bytes
    """

    if lineterminator is not None and len(lineterminator) == 1:
        kwargs["lineterminator"] = lineterminator
    else:
        lineterminator = "\n"

    if compression == "infer":
        paths = get_fs_token_paths(urlpath, mode="rb", storage_options=storage_options)[2]
        compression = infer_compression(paths[0])

    if isinstance(blocksize, str):
        blocksize = parse_bytes(blocksize)
    if blocksize and compression:
        blocksize = None

    b_out = read_bytes(
        urlpath,
        delimiter=lineterminator.encode(),
        blocksize=blocksize,
        sample=False,
        compression=compression,
        include_path=True,
        **(storage_options or {}),
    )
    _, values, paths = b_out

    if not isinstance(values[0], (tuple, list)):
        values = [values]

    return paths, [len(v) for v in values]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tensorflow_metadata/proto/v0/path.proto, tensorflow_metadata/proto/v0/schema.proto
# plugin: python-betterproto
import json
import os
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Dict, List

import betterproto

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"
from google.protobuf import json_format  # noqa: E402
from google.protobuf.any_pb2 import Any as AnyPb2  # noqa: E402
from google.protobuf.struct_pb2 import ListValue, Struct  # noqa: E402


# This is a manual hack to make it possible to parse the Any types in annotations
@dataclass(eq=False, repr=False)
class Any(betterproto.Message):
    type_url: str = betterproto.string_field(1)
    value: betterproto.Message = betterproto.message_field(2)

    def from_dict(self, value: dict):
        if "@type" in value:
            self.value = value["value"]
            self.type_url = value["@type"]
        else:
            msg_struct = Struct()
            any_pack = AnyPb2()
            json_formatted = json_format.ParseDict(value, msg_struct)
            any_pack.Pack(json_formatted)
            self.value = any_pack.value
            self.type_url = any_pack.type_url
        return self

    def to_dict(
        self,
        casing: betterproto.Casing = betterproto.Casing.CAMEL,
        include_default_values: bool = False,
    ):
        if isinstance(self.value, dict):
            ret = self.value
        else:
            msg_struct = Struct()
            msg_struct.ParseFromString(self.value)
            ret = dict(msg_struct.items())

        # the ParseFromString method will return Struct types for embedded
        # dictionaries, and ListValue for embedded lists.
        # Recursively convert embedded values
        def _ensure_value(v):
            if isinstance(v, dict):
                return _ensure_dict(v)
            elif isinstance(v, (list, ListValue)):
                return [_ensure_value(val) for val in v]
            elif isinstance(v, Struct):
                return _ensure_dict(dict(v.items()))
            return v

        def _ensure_dict(x):
            return {k: _ensure_value(v) for k, v in x.items()}

        return _ensure_dict(ret)


class LifecycleStage(betterproto.Enum):
    """
    LifecycleStage. Only UNKNOWN_STAGE, BETA, and PRODUCTION features are
    actually validated. PLANNED, ALPHA, DISABLED, and DEBUG are treated as
    DEPRECATED.
    """

    UNKNOWN_STAGE = 0
    PLANNED = 1
    ALPHA = 2
    BETA = 3
    PRODUCTION = 4
    DEPRECATED = 5
    DEBUG_ONLY = 6
    DISABLED = 7


class FeatureType(betterproto.Enum):
    """
    Describes the physical representation of a feature. It may be different
    than the logical representation, which is represented as a Domain.
    """

    TYPE_UNKNOWN = 0
    BYTES = 1
    INT = 2
    FLOAT = 3
    STRUCT = 4


class TimeDomainIntegerTimeFormat(betterproto.Enum):
    FORMAT_UNKNOWN = 0
    UNIX_DAYS = 5
    UNIX_SECONDS = 1
    UNIX_MILLISECONDS = 2
    UNIX_MICROSECONDS = 3
    UNIX_NANOSECONDS = 4


class TimeOfDayDomainIntegerTimeOfDayFormat(betterproto.Enum):
    FORMAT_UNKNOWN = 0
    PACKED_64_NANOS = 1


class TensorRepresentationRowPartitionDType(betterproto.Enum):
    UNSPECIFIED = 0
    INT64 = 1
    INT32 = 2


@dataclass
class Path(betterproto.Message):
    """
    A path is a more general substitute for the name of a field or feature that
    can be used for flat examples as well as structured data. For example, if
    we had data in a protocol buffer: message Person {   int age = 1;
    optional string gender = 2;   repeated Person parent = 3; } Thus, here the
    path {step:["parent", "age"]} in statistics would refer to the age of a
    parent, and {step:["parent", "parent", "age"]} would refer to the age of a
    grandparent. This allows us to distinguish between the statistics of
    parents' ages and grandparents' ages. In general, repeated messages are to
    be preferred to linked lists of arbitrary length. For SequenceExample, if
    we have a feature list "foo", this is represented by {step:["##SEQUENCE##",
    "foo"]}.
    """

    # Any string is a valid step. However, whenever possible have a step be
    # [A-Za-z0-9_]+.
    step: List[str] = betterproto.string_field(1)


@dataclass
class Schema(betterproto.Message):
    """Message to represent schema information. NextID: 14"""

    # Features described in this schema.
    feature: List["Feature"] = betterproto.message_field(1)
    # Sparse features described in this schema.
    sparse_feature: List["SparseFeature"] = betterproto.message_field(6)
    # Weighted features described in this schema.
    weighted_feature: List["WeightedFeature"] = betterproto.message_field(12)
    # declared as top-level features in <feature>. String domains referenced in
    # the features.
    string_domain: List["StringDomain"] = betterproto.message_field(4)
    # top level float domains that can be reused by features
    float_domain: List["FloatDomain"] = betterproto.message_field(9)
    # top level int domains that can be reused by features
    int_domain: List["IntDomain"] = betterproto.message_field(10)
    # Default environments for each feature. An environment represents both a
    # type of location (e.g. a server or phone) and a time (e.g. right before
    # model X is run). In the standard scenario, 99% of the features should be in
    # the default environments TRAINING, SERVING, and the LABEL (or labels) AND
    # WEIGHT is only available at TRAINING (not at serving). Other possible
    # variations: 1. There may be TRAINING_MOBILE, SERVING_MOBILE,
    # TRAINING_SERVICE,    and SERVING_SERVICE. 2. If one is ensembling three
    # models, where the models of the first    three models are available
    # for the ensemble model, there may be    TRAINING, SERVING_INITIAL,
    # SERVING_ENSEMBLE. See FeatureProto::not_in_environment and
    # FeatureProto::in_environment.
    default_environment: List[str] = betterproto.string_field(5)
    # Additional information about the schema as a whole. Features may also be
    # annotated individually.
    annotation: "Annotation" = betterproto.message_field(8)
    # Dataset-level constraints. This is currently used for specifying
    # information about changes in num_examples.
    dataset_constraints: "DatasetConstraints" = betterproto.message_field(11)
    # TensorRepresentation groups. The keys are the names of the groups. Key ""
    # (empty string) denotes the "default" group, which is what should be used
    # when a group name is not provided. See the documentation at
    # TensorRepresentationGroup for more info. Under development. DO NOT USE.
    tensor_representation_group: Dict[str, "TensorRepresentationGroup"] = betterproto.map_field(
        13, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass
class ValueCountList(betterproto.Message):
    value_count: List["ValueCount"] = betterproto.message_field(1)


@dataclass
class Feature(betterproto.Message):
    """
    Describes schema-level information about a specific feature. NextID: 33
    """

    # The name of the feature.
    name: str = betterproto.string_field(1)
    # This field is no longer supported. Instead, use: lifecycle_stage:
    # DEPRECATED TODO(b/111450258): remove this.
    deprecated: bool = betterproto.bool_field(2)
    # Constraints on the presence of this feature in the examples.
    presence: "FeaturePresence" = betterproto.message_field(14, group="presence_constraints")
    # Only used in the context of a "group" context, e.g., inside a sequence.
    group_presence: "FeaturePresenceWithinGroup" = betterproto.message_field(
        17, group="presence_constraints"
    )
    # The feature has a fixed shape corresponding to a multi-dimensional tensor.
    shape: "FixedShape" = betterproto.message_field(23, group="shape_type")
    # The feature doesn't have a well defined shape. All we know are limits on
    # the minimum and maximum number of values.
    value_count: "ValueCount" = betterproto.message_field(5, group="shape_type")
    # Captures the same information as value_count but for features with nested
    # values. A ValueCount is provided for each nest level.
    value_counts: "ValueCountList" = betterproto.message_field(32, group="shape_type")
    # Physical type of the feature's values. Note that you can have: type: BYTES
    # int_domain: {   min: 0   max: 3 } This would be a field that is
    # syntactically BYTES (i.e. strings), but semantically an int, i.e. it would
    # be "0", "1", "2", or "3".
    type: "FeatureType" = betterproto.enum_field(6)
    # Reference to a domain defined at the schema level.
    domain: str = betterproto.string_field(7, group="domain_info")
    # Inline definitions of domains.
    int_domain: "IntDomain" = betterproto.message_field(9, group="domain_info")
    float_domain: "FloatDomain" = betterproto.message_field(10, group="domain_info")
    string_domain: "StringDomain" = betterproto.message_field(11, group="domain_info")
    bool_domain: "BoolDomain" = betterproto.message_field(13, group="domain_info")
    struct_domain: "StructDomain" = betterproto.message_field(29, group="domain_info")
    # Supported semantic domains.
    natural_language_domain: "NaturalLanguageDomain" = betterproto.message_field(
        24, group="domain_info"
    )
    image_domain: "ImageDomain" = betterproto.message_field(25, group="domain_info")
    mid_domain: "MIDDomain" = betterproto.message_field(26, group="domain_info")
    url_domain: "URLDomain" = betterproto.message_field(27, group="domain_info")
    time_domain: "TimeDomain" = betterproto.message_field(28, group="domain_info")
    time_of_day_domain: "TimeOfDayDomain" = betterproto.message_field(30, group="domain_info")
    # Constraints on the distribution of the feature values. Only supported for
    # StringDomains.
    distribution_constraints: "DistributionConstraints" = betterproto.message_field(15)
    # Additional information about the feature for documentation purpose.
    annotation: "Annotation" = betterproto.message_field(16)
    # Tests comparing the distribution to the associated serving data.
    skew_comparator: "FeatureComparator" = betterproto.message_field(18)
    # Tests comparing the distribution between two consecutive spans (e.g. days).
    drift_comparator: "FeatureComparator" = betterproto.message_field(21)
    # List of environments this feature is present in. Should be disjoint from
    # not_in_environment. This feature is in environment "foo" if: ("foo" is in
    # in_environment or default_environment) AND "foo" is not in
    # not_in_environment. See Schema::default_environment.
    in_environment: List[str] = betterproto.string_field(20)
    # List of environments this feature is not present in. Should be disjoint
    # from of in_environment. See Schema::default_environment and in_environment.
    not_in_environment: List[str] = betterproto.string_field(19)
    # The lifecycle stage of a feature. It can also apply to its descendants.
    # i.e., if a struct is DEPRECATED, its children are implicitly deprecated.
    lifecycle_stage: "LifecycleStage" = betterproto.enum_field(22)
    # Constraints on the number of unique values for a given feature. This is
    # supported for string and categorical features only.
    unique_constraints: "UniqueConstraints" = betterproto.message_field(31)


@dataclass
class Annotation(betterproto.Message):
    """Additional information about the schema or about a feature."""

    # Tags can be used to mark features. For example, tag on user_age feature can
    # be `user_feature`, tag on user_country feature can be `location_feature`,
    # `user_feature`.
    tag: List[str] = betterproto.string_field(1)
    # Free-text comments. This can be used as a description of the feature,
    # developer notes etc.
    comment: List[str] = betterproto.string_field(2)
    # Application-specific metadata may be attached here.
    extra_metadata: List[Any] = betterproto.message_field(3)

    @property
    def metadata(self):
        if self.comment:
            metadata = SimpleNamespace(**json.loads(self.comment[-1]))
            return metadata

        return None


@dataclass
class NumericValueComparator(betterproto.Message):
    """
    Checks that the ratio of the current value to the previous value is not
    below the min_fraction_threshold or above the max_fraction_threshold. That
    is, previous value * min_fraction_threshold <= current value <= previous
    value * max_fraction_threshold. To specify that the value cannot change,
    set both min_fraction_threshold and max_fraction_threshold to 1.0.
    """

    min_fraction_threshold: float = betterproto.double_field(1)
    max_fraction_threshold: float = betterproto.double_field(2)


@dataclass
class DatasetConstraints(betterproto.Message):
    """Constraints on the entire dataset."""

    # Tests differences in number of examples between the current data and the
    # previous span.
    num_examples_drift_comparator: "NumericValueComparator" = betterproto.message_field(1)
    # Tests comparisons in number of examples between the current data and the
    # previous version of that data.
    num_examples_version_comparator: "NumericValueComparator" = betterproto.message_field(2)
    # Minimum number of examples in the dataset.
    min_examples_count: int = betterproto.int64_field(3)
    # Maximum number of examples in the dataset.
    max_examples_count: int = betterproto.int64_field(4)


@dataclass
class FixedShape(betterproto.Message):
    """
    Specifies a fixed shape for the feature's values. The immediate implication
    is that each feature has a fixed number of values. Moreover, these values
    can be parsed in a multi-dimensional tensor using the specified axis sizes.
    The FixedShape defines a lexicographical ordering of the data. For
    instance, if there is a FixedShape {   dim {size:3} dim {size:2} } then
    tensor[0][0]=field[0] then tensor[0][1]=field[1] then tensor[1][0]=field[2]
    then tensor[1][1]=field[3] then tensor[2][0]=field[4] then
    tensor[2][1]=field[5] The FixedShape message is identical with the
    TensorFlow TensorShape proto message.
    """

    # The dimensions that define the shape. The total number of values in each
    # example is the product of sizes of each dimension.
    dim: List["FixedShapeDim"] = betterproto.message_field(2)


@dataclass
class FixedShapeDim(betterproto.Message):
    """An axis in a multi-dimensional feature representation."""

    size: int = betterproto.int64_field(1)
    # Optional name of the tensor dimension.
    name: str = betterproto.string_field(2)


@dataclass
class ValueCount(betterproto.Message):
    """
    Limits on maximum and minimum number of values in a single example (when
    the feature is present). Use this when the minimum value count can be
    different than the maximum value count. Otherwise prefer FixedShape.
    """

    min: int = betterproto.int64_field(1)
    max: int = betterproto.int64_field(2)


@dataclass
class WeightedFeature(betterproto.Message):
    """
    Represents a weighted feature that is encoded as a combination of raw base
    features. The `weight_feature` should be a float feature with identical
    shape as the `feature`. This is useful for representing weights associated
    with categorical tokens (e.g. a TFIDF weight associated with each token).
    TODO(b/142122960): Handle WeightedCategorical end to end in TFX
    (validation, TFX Unit Testing, etc)
    """

    # Name for the weighted feature. This should not clash with other features in
    # the same schema.
    name: str = betterproto.string_field(1)
    # Path of a base feature to be weighted. Required.
    feature: "Path" = betterproto.message_field(2)
    # Path of weight feature to associate with the base feature. Must be same
    # shape as feature. Required.
    weight_feature: "Path" = betterproto.message_field(3)
    # The lifecycle_stage determines where a feature is expected to be used, and
    # therefore how important issues with it are.
    lifecycle_stage: "LifecycleStage" = betterproto.enum_field(4)


@dataclass
class SparseFeature(betterproto.Message):
    """
    A sparse feature represents a sparse tensor that is encoded with a
    combination of raw features, namely index features and a value feature.
    Each index feature defines a list of indices in a different dimension.
    """

    # Name for the sparse feature. This should not clash with other features in
    # the same schema.
    name: str = betterproto.string_field(1)
    # This field is no longer supported. Instead, use: lifecycle_stage:
    # DEPRECATED TODO(b/111450258): remove this.
    deprecated: bool = betterproto.bool_field(2)
    # The lifecycle_stage determines where a feature is expected to be used, and
    # therefore how important issues with it are.
    lifecycle_stage: "LifecycleStage" = betterproto.enum_field(7)
    # Constraints on the presence of this feature in examples. Deprecated, this
    # is inferred by the referred features.
    presence: "FeaturePresence" = betterproto.message_field(4)
    # Shape of the sparse tensor that this SparseFeature represents. Currently
    # not supported. TODO(b/109669962): Consider deriving this from the referred
    # features.
    dense_shape: "FixedShape" = betterproto.message_field(5)
    # Features that represent indexes. Should be integers >= 0.
    index_feature: List["SparseFeatureIndexFeature"] = betterproto.message_field(6)
    # If true then the index values are already sorted lexicographically.
    is_sorted: bool = betterproto.bool_field(8)
    value_feature: "SparseFeatureValueFeature" = betterproto.message_field(9)
    # Type of value feature. Deprecated, this is inferred by the referred
    # features.
    type: "FeatureType" = betterproto.enum_field(10)


@dataclass
class SparseFeatureIndexFeature(betterproto.Message):
    # Name of the index-feature. This should be a reference to an existing
    # feature in the schema.
    name: str = betterproto.string_field(1)


@dataclass
class SparseFeatureValueFeature(betterproto.Message):
    # Name of the value-feature. This should be a reference to an existing
    # feature in the schema.
    name: str = betterproto.string_field(1)


@dataclass
class DistributionConstraints(betterproto.Message):
    """
    Models constraints on the distribution of a feature's values.
    TODO(martinz): replace min_domain_mass with max_off_domain (but slowly).
    """

    # The minimum fraction (in [0,1]) of values across all examples that should
    # come from the feature's domain, e.g.:   1.0  => All values must come from
    # the domain.    .9  => At least 90% of the values must come from the domain.
    min_domain_mass: float = betterproto.double_field(1)


@dataclass
class FeatureCoverageConstraints(betterproto.Message):
    """Encodes vocabulary coverage constraints."""

    # Fraction of feature values that map to a vocab entry (i.e. are not oov).
    min_coverage: float = betterproto.float_field(1)
    # Average length of tokens. Used for cases such as wordpiece that fallback to
    # character-level tokenization.
    min_avg_token_length: float = betterproto.float_field(2)
    # String tokens to exclude when calculating min_coverage and
    # min_avg_token_length. Useful for tokens such as [PAD].
    excluded_string_tokens: List[str] = betterproto.string_field(3)
    # Integer tokens to exclude when calculating min_coverage and
    # min_avg_token_length.
    excluded_int_tokens: List[int] = betterproto.int64_field(4)
    # String tokens to treat as oov tokens (e.g. [UNK]). These tokens are also
    # excluded when calculating avg token length.
    oov_string_tokens: List[str] = betterproto.string_field(5)


@dataclass
class SequenceValueConstraints(betterproto.Message):
    """Encodes constraints on specific values in sequences."""

    int_value: int = betterproto.int64_field(1, group="value")
    string_value: str = betterproto.string_field(2, group="value")
    # Min / max number of times the value can occur in a sequence.
    min_per_sequence: int = betterproto.int64_field(3)
    max_per_sequence: int = betterproto.int64_field(4)
    # Min / max fraction of sequences that must contain the value.
    min_fraction_of_sequences: float = betterproto.float_field(5)
    max_fraction_of_sequences: float = betterproto.float_field(6)


@dataclass
class SequenceLengthConstraints(betterproto.Message):
    """Encodes constraints on sequence lengths."""

    # Token values (int and string) that are excluded when calculating sequence
    # length.
    excluded_int_value: List[int] = betterproto.int64_field(1)
    excluded_string_value: List[str] = betterproto.string_field(2)
    # Min / max sequence length.
    min_sequence_length: int = betterproto.int64_field(3)
    max_sequence_length: int = betterproto.int64_field(4)


@dataclass
class IntDomain(betterproto.Message):
    """
    Encodes information for domains of integer values. Note that FeatureType
    could be either INT or BYTES.
    """

    # Id of the domain. Required if the domain is defined at the schema level. If
    # so, then the name must be unique within the schema.
    name: str = betterproto.string_field(1)
    # Min and max values for the domain.
    min: int = betterproto.int64_field(3)
    max: int = betterproto.int64_field(4)
    # If true then the domain encodes categorical values (i.e., ids) rather than
    # ordinal values.
    is_categorical: bool = betterproto.bool_field(5)


@dataclass
class FloatDomain(betterproto.Message):
    """
    Encodes information for domains of float values. Note that FeatureType
    could be either INT or BYTES.
    """

    # Id of the domain. Required if the domain is defined at the schema level. If
    # so, then the name must be unique within the schema.
    name: str = betterproto.string_field(1)
    # Min and max values of the domain.
    min: float = betterproto.float_field(3)
    max: float = betterproto.float_field(4)
    # If true, feature should not contain NaNs.
    disallow_nan: bool = betterproto.bool_field(5)
    # If true, feature should not contain Inf or -Inf.
    disallow_inf: bool = betterproto.bool_field(6)
    # If True, this indicates that the feature is semantically an embedding. This
    # can be useful for distinguishing fixed dimensional numeric features that
    # should be fed to a prediction unmodified.
    is_embedding: bool = betterproto.bool_field(7)


@dataclass
class StructDomain(betterproto.Message):
    """
    Domain for a recursive struct. NOTE: If a feature with a StructDomain is
    deprecated, then all the child features (features and sparse_features of
    the StructDomain) are also considered to be deprecated.  Similarly child
    features can only be in environments of the parent feature.
    """

    feature: List["Feature"] = betterproto.message_field(1)
    sparse_feature: List["SparseFeature"] = betterproto.message_field(2)


@dataclass
class StringDomain(betterproto.Message):
    """Encodes information for domains of string values."""

    # Id of the domain. Required if the domain is defined at the schema level. If
    # so, then the name must be unique within the schema.
    name: str = betterproto.string_field(1)
    # The values appearing in the domain.
    value: List[str] = betterproto.string_field(2)


@dataclass
class BoolDomain(betterproto.Message):
    """
    Encodes information about the domain of a boolean attribute that encodes
    its TRUE/FALSE values as strings, or 0=false, 1=true. Note that FeatureType
    could be either INT or BYTES.
    """

    # Id of the domain. Required if the domain is defined at the schema level. If
    # so, then the name must be unique within the schema.
    name: str = betterproto.string_field(1)
    # Strings values for TRUE/FALSE.
    true_value: str = betterproto.string_field(2)
    false_value: str = betterproto.string_field(3)


@dataclass
class NaturalLanguageDomain(betterproto.Message):
    """Natural language text."""

    # Name of the vocabulary associated with the NaturalLanguageDomain. When
    # computing and validating stats using TFDV, tfdv.StatsOptions.vocab_paths
    # should map this name to a vocabulary file.
    vocabulary: str = betterproto.string_field(1)
    coverage: "FeatureCoverageConstraints" = betterproto.message_field(2)
    token_constraints: List["SequenceValueConstraints"] = betterproto.message_field(3)
    sequence_length_constraints: "SequenceLengthConstraints" = betterproto.message_field(5)
    # Specifies the location constraints as a function of the tokens specified in
    # token_constraints. String tokens will be specified by S_TOKEN_, (e.g.
    # S_(PAD)_) and integer tokens will be specified as I_#_ (e.g. I_123_). A_T_
    # will match any token that has not been specified in token_constraints.
    # Parenthesis, +, and * are supported. _ will be escapable with a \ for
    # tokens containing it (e.g. FOO\_BAR). For example, a two-sequence BERT
    # model may look as follows: S_(CLS)_ A_T_+ S_(SEP)_ A_T_+ S_(SEP)_ S_(PAD)_*
    # Note: Support for this field is not yet implemented. Please do not use.
    # TODO(b/188095987): Remove warning once field is implemented.
    location_constraint_regex: str = betterproto.string_field(4)


@dataclass
class ImageDomain(betterproto.Message):
    """Image data."""

    # If set, at least this fraction of values should be TensorFlow supported
    # images.
    minimum_supported_image_fraction: float = betterproto.float_field(1)
    # If set, image should have less than this value of undecoded byte size.
    max_image_byte_size: int = betterproto.int64_field(2)


@dataclass
class MIDDomain(betterproto.Message):
    """Knowledge graph ID, see: https://www.wikidata.org/wiki/Property:P646"""


@dataclass
class URLDomain(betterproto.Message):
    """A URL, see: https://en.wikipedia.org/wiki/URL"""


@dataclass
class TimeDomain(betterproto.Message):
    """Time or date representation."""

    # Expected format that contains a combination of regular characters and
    # special format specifiers. Format specifiers are a subset of the strptime
    # standard.
    string_format: str = betterproto.string_field(1, group="format")
    # Expected format of integer times.
    integer_format: "TimeDomainIntegerTimeFormat" = betterproto.enum_field(2, group="format")


@dataclass
class TimeOfDayDomain(betterproto.Message):
    """Time of day, without a particular date."""

    # Expected format that contains a combination of regular characters and
    # special format specifiers. Format specifiers are a subset of the strptime
    # standard.
    string_format: str = betterproto.string_field(1, group="format")
    # Expected format of integer times.
    integer_format: "TimeOfDayDomainIntegerTimeOfDayFormat" = betterproto.enum_field(
        2, group="format"
    )


@dataclass
class FeaturePresence(betterproto.Message):
    """Describes constraints on the presence of the feature in the data."""

    # Minimum fraction of examples that have this feature.
    min_fraction: float = betterproto.double_field(1)
    # Minimum number of examples that have this feature.
    min_count: int = betterproto.int64_field(2)


@dataclass
class FeaturePresenceWithinGroup(betterproto.Message):
    """
    Records constraints on the presence of a feature inside a "group" context
    (e.g., .presence inside a group of features that define a sequence).
    """

    required: bool = betterproto.bool_field(1)


@dataclass
class InfinityNorm(betterproto.Message):
    """
    Checks that the L-infinity norm is below a certain threshold between the
    two discrete distributions. Since this is applied to a
    FeatureNameStatistics, it only considers the top k. L_infty(p,q) = max_i
    |p_i-q_i|
    """

    # The InfinityNorm is in the interval [0.0, 1.0] so sensible bounds should be
    # in the interval [0.0, 1.0).
    threshold: float = betterproto.double_field(1)


@dataclass
class JensenShannonDivergence(betterproto.Message):
    """
    Checks that the approximate Jensen-Shannon Divergence is below a certain
    threshold between the two distributions.
    """

    # The JensenShannonDivergence will be in the interval [0.0, 1.0] so sensible
    # bounds should be in the interval [0.0, 1.0).
    threshold: float = betterproto.double_field(1)


@dataclass
class FeatureComparator(betterproto.Message):
    infinity_norm: "InfinityNorm" = betterproto.message_field(1)
    jensen_shannon_divergence: "JensenShannonDivergence" = betterproto.message_field(2)


@dataclass
class UniqueConstraints(betterproto.Message):
    """
    Checks that the number of unique values is greater than or equal to the
    min, and less than or equal to the max.
    """

    min: int = betterproto.int64_field(1)
    max: int = betterproto.int64_field(2)


@dataclass
class TensorRepresentation(betterproto.Message):
    """
    A TensorRepresentation captures the intent for converting columns in a
    dataset to TensorFlow Tensors (or more generally, tf.CompositeTensors).
    Note that one tf.CompositeTensor may consist of data from multiple columns,
    for example, a N-dimensional tf.SparseTensor may need N + 1 columns to
    provide the sparse indices and values. Note that the "column name" that a
    TensorRepresentation needs is a string, not a Path -- it means that the
    column name identifies a top-level Feature in the schema (i.e. you cannot
    specify a Feature nested in a STRUCT Feature).
    """

    dense_tensor: "TensorRepresentationDenseTensor" = betterproto.message_field(1, group="kind")
    varlen_sparse_tensor: "TensorRepresentationVarLenSparseTensor" = betterproto.message_field(
        2, group="kind"
    )
    sparse_tensor: "TensorRepresentationSparseTensor" = betterproto.message_field(3, group="kind")
    ragged_tensor: "TensorRepresentationRaggedTensor" = betterproto.message_field(4, group="kind")


@dataclass
class TensorRepresentationDefaultValue(betterproto.Message):
    float_value: float = betterproto.double_field(1, group="kind")
    # Note that the data column might be of a shorter integral type. It's the
    # user's responsitiblity to make sure the default value fits that type.
    int_value: int = betterproto.int64_field(2, group="kind")
    bytes_value: bytes = betterproto.bytes_field(3, group="kind")
    # uint_value should only be used if the default value can't fit in a int64
    # (`int_value`).
    uint_value: int = betterproto.uint64_field(4, group="kind")


@dataclass
class TensorRepresentationDenseTensor(betterproto.Message):
    """A tf.Tensor"""

    # Identifies the column in the dataset that provides the values of this
    # Tensor.
    column_name: str = betterproto.string_field(1)
    # The shape of each row of the data (i.e. does not include the batch
    # dimension)
    shape: "FixedShape" = betterproto.message_field(2)
    # If this column is missing values in a row, the default_value will be used
    # to fill that row.
    default_value: "TensorRepresentationDefaultValue" = betterproto.message_field(3)


@dataclass
class TensorRepresentationVarLenSparseTensor(betterproto.Message):
    """A ragged tf.SparseTensor that models nested lists."""

    # Identifies the column in the dataset that should be converted to the
    # VarLenSparseTensor.
    column_name: str = betterproto.string_field(1)


@dataclass
class TensorRepresentationSparseTensor(betterproto.Message):
    """
    A tf.SparseTensor whose indices and values come from separate data columns.
    This will replace Schema.sparse_feature eventually. The index columns must
    be of INT type, and all the columns must co-occur and have the same valency
    at the same row.
    """

    # The dense shape of the resulting SparseTensor (does not include the batch
    # dimension).
    dense_shape: "FixedShape" = betterproto.message_field(1)
    # The columns constitute the coordinates of the values. indices_column[i][j]
    # contains the coordinate of the i-th dimension of the j-th value.
    index_column_names: List[str] = betterproto.string_field(2)
    # The column that contains the values.
    value_column_name: str = betterproto.string_field(3)


@dataclass
class TensorRepresentationRaggedTensor(betterproto.Message):
    """
    A tf.RaggedTensor that models nested lists. Currently there is no way for
    the user to specify the shape of the leaf value (the innermost value tensor
    of the RaggedTensor). The leaf value will always be a 1-D tensor.
    """

    # Identifies the leaf feature that provides values of the RaggedTensor.
    # struct type sub fields. The first step of the path refers to a top-level
    # feature in the data. The remaining steps refer to STRUCT features under the
    # top-level feature, recursively. If the feature has N outer ragged lists,
    # they will become the first N dimensions of the resulting RaggedTensor and
    # the contents will become the flat_values.
    feature_path: "Path" = betterproto.message_field(1)
    # The result RaggedTensor would be of shape: [B, D_0, D_1, ..., D_N, P_0,
    # P_1, ..., P_M, U_0, U_1, ..., U_P] Where the dimensions belong to different
    # categories: * B: Batch size dimension * D_n: Dimensions specified by the
    # nested structure specified by the value path until the leaf node. n>=1. *
    # P_m: Dimensions specified by the partitions that do not define any fixed
    # diomension size. m>=0. * U_0: Dimensions specified by the latest partitions
    # of type uniform_row_length that can define the fixed inner shape of the
    # tensor. If iterationg the partitions from the end to the beginning, these
    # dimensions are defined by all the continuous uniform_row_length partitions
    # present. p>=0.
    partition: List["TensorRepresentationRaggedTensorPartition"] = betterproto.message_field(3)
    # The data type of the ragged tensor's row partitions. This will default to
    # INT64 if it is not specified.
    row_partition_dtype: "TensorRepresentationRowPartitionDType" = betterproto.enum_field(2)


@dataclass
class TensorRepresentationRaggedTensorPartition(betterproto.Message):
    """Further partition of the feature values at the leaf level."""

    # If the final element(s) of partition are uniform_row_lengths [U0, U1, ...]
    # , then the result RaggedTensor will have  their flat values (a dense
    # tensor) being of shape [U0, U1, ...]. Otherwise, a uniform_row_length
    # simply means a ragged dimension with row_lengths
    # [uniform_row_length]*nrows.
    uniform_row_length: int = betterproto.int64_field(1, group="kind")
    # Identifies a leaf feature who share the same parent of value_feature_path
    # that contains the partition row lengths.
    row_length: str = betterproto.string_field(2, group="kind")


@dataclass
class TensorRepresentationGroup(betterproto.Message):
    """
    A TensorRepresentationGroup is a collection of TensorRepresentations with
    names. These names may serve as identifiers when converting the dataset to
    a collection of Tensors or tf.CompositeTensors. For example, given the
    following group: {   key: "dense_tensor"   tensor_representation {
    dense_tensor {       column_name: "univalent_feature"       shape {
    dim {           size: 1         }       }       default_value {
    float_value: 0       }     }   } } {   key: "varlen_sparse_tensor"
    tensor_representation {     varlen_sparse_tensor {       column_name:
    "multivalent_feature"     }   } } Then the schema is expected to have
    feature "univalent_feature" and "multivalent_feature", and when a batch of
    data is converted to Tensors using this TensorRepresentationGroup, the
    result may be the following dict: {   "dense_tensor": tf.Tensor(...),
    "varlen_sparse_tensor": tf.SparseTensor(...), }
    """

    tensor_representation: Dict[str, "TensorRepresentation"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import gzip
import importlib
import os
import shutil
import tarfile
import urllib.request
import warnings
import zipfile
from contextvars import ContextVar

import dask
from dask.dataframe.optimize import optimize as dd_optimize
from dask.distributed import Client, get_client
from tqdm import tqdm

_merlin_dask_client = ContextVar("_merlin_dask_client", default="auto")

try:
    from numba import cuda
except ImportError:
    cuda = None

try:
    import psutil
except ImportError:
    psutil = None


def pynvml_mem_size(kind="total", index=0):
    import pynvml

    pynvml.nvmlInit()
    size = None
    if kind == "free":
        size = int(pynvml.nvmlDeviceGetMemoryInfo(pynvml.nvmlDeviceGetHandleByIndex(index)).free)
    elif kind == "total":
        size = int(pynvml.nvmlDeviceGetMemoryInfo(pynvml.nvmlDeviceGetHandleByIndex(index)).total)
    else:
        raise ValueError("{0} not a supported option for device_mem_size.".format(kind))
    pynvml.nvmlShutdown()
    return size


def device_mem_size(kind="total", cpu=False):

    # Use psutil (if available) for cpu mode
    if cpu and psutil:
        if kind == "total":
            return psutil.virtual_memory().total
        elif kind == "free":
            return psutil.virtual_memory().free
    elif cpu:
        warnings.warn("Please install psutil for full cpu=True support.")
        # Assume 1GB of memory
        return int(1e9)

    if kind not in ["free", "total"]:
        raise ValueError("{0} not a supported option for device_mem_size.".format(kind))
    try:
        if kind == "free":
            return int(cuda.current_context().get_memory_info()[0])
        else:
            return int(cuda.current_context().get_memory_info()[1])
    except NotImplementedError:
        if kind == "free":
            # Not using NVML "free" memory, because it will not include RMM-managed memory
            warnings.warn("get_memory_info is not supported. Using total device memory from NVML.")
        size = pynvml_mem_size(kind="total", index=0)
    return size


def get_rmm_size(size):
    return (size // 256) * 256


def download_file(url, local_filename, unzip_files=True, redownload=True):
    """utility function to download a dataset file (movielens/criteo/rossmann etc)
    locally, displaying a progress bar during download"""
    local_filename = os.path.abspath(local_filename)
    path = os.path.dirname(local_filename)
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)

    if not url.startswith("http"):
        raise ValueError(f"Unhandled url scheme on {url} - this function only is for http")

    if redownload or not os.path.exists(local_filename):
        desc = f"downloading {os.path.basename(local_filename)}"
        with tqdm(unit="B", unit_scale=True, desc=desc) as progress:

            def report(chunk, chunksize, total):
                if not progress.total:
                    progress.reset(total=total)
                progress.update(chunksize)

            opener = urllib.request.build_opener()
            opener.addheaders = [("Accept-Encoding", "gzip, deflate"), ("Accept", "*/*")]

            urllib.request.install_opener(opener)
            urllib.request.urlretrieve(url, local_filename, reporthook=report)  # nosec

    if unzip_files and local_filename.endswith(".zip"):
        with zipfile.ZipFile(local_filename) as z:
            for filename in tqdm(z.infolist(), desc="unzipping files", unit="files"):
                z.extract(filename, path)

    elif unzip_files and local_filename.endswith(".tgz"):
        with tarfile.open(local_filename, "r") as tar:
            for filename in tqdm(tar.getnames(), desc="untarring files", unit="files"):
                tar.extract(filename, path)

    elif unzip_files and local_filename.endswith(".gz"):
        with gzip.open(local_filename, "rb") as input_file:
            with open(local_filename[:-3], "wb") as output_file:
                shutil.copyfileobj(input_file, output_file)


def ensure_optimize_dataframe_graph(ddf=None, dsk=None, keys=None):
    # Perform HLG DataFrame optimizations
    #
    # If `ddf` is specified, an optimized Dataframe
    # collection will be returned. If `dsk` and `keys`
    # are specified, an optimized graph will be returned.
    #
    # These optimizations are performed automatically
    # when a DataFrame collection is computed/persisted,
    # but they are NOT always performed when statistics
    # are computed. The purpose of this utility is to
    # ensure that the Dataframe-based optimizations are
    # always applied.

    if ddf is None:
        if dsk is None or keys is None:
            raise ValueError("Must specify both `dsk` and `keys` if `ddf` is not supplied.")
    dsk = ddf.dask if dsk is None else dsk
    keys = ddf.__dask_keys__() if keys is None else keys

    if isinstance(dsk, dask.highlevelgraph.HighLevelGraph):
        with dask.config.set({"optimization.fuse.active": False}):
            dsk = dd_optimize(dsk, keys=keys)

    if ddf is None:
        # Return optimized graph
        return dsk

    # Return optimized ddf
    ddf.dask = dsk
    return ddf


class Distributed:
    """Distributed-Execution Context Manager

    The purpose of this execution-manager utility is to
    provide an intuitive context manager for distributed
    (multi-GPU/CPU) scheduling and execution with Dask.

    NOTE: For multi-node execution, it is the users
    responsibility to create a distributed Dask cluster
    with an appropriate deployment technology.  This
    class only supports the automatic generation of
    local (single-machine) clusters. However, the class
    can be used to connect to any existing dask cluster
    (local or not), as long as a valid `client` argument
    can be defined.

    Parameters
    -----------
    client : `dask.distributed.Client`; Optional
        The client to use for distributed-Dask execution.
    new_cluster : {"cuda", "cpu", None}
        Type of local cluster to generate in the case that a
        global client is not detected (or `force_new=True`).
        "cuda" corresponds to `dask_cuda.LocalCUDACluster`,
        while "cpu" corresponds to `distributed.LocalCluster`.
        Default is "cuda" if GPU support is detected.
    force_new : bool
        Whether to force the creation of a new local cluster
        in the case that a global client object is already
        detected. Default is False.
    **cluster_options :
        Key-word arguments to pass to the local-cluster
        constructor specified by `new_cluster` (e.g.
        `n_workers=2`).

    Examples
    --------
    The easiest way to use `Distributed` is within a
    conventional `with` statement::

        from merlin.core.utils import Disributed

        workflow = nvt.Workflow(["col"] >> ops.Normalize())
        dataset = nvt.Dataset(...)
        with Distributed():
            workflow.transform(dataset).to_parquet(...)

    In this case, all Dask-based scheduling and execution
    required within the `with Distributed()` block will be
    performed using a distributed cluster. If an existing
    client is not detected, a default `LocalCUDACluster`
    or `LocalCluster` will be automatically deployed (the
    specific type depending on GPU support).

    Alternatively, the distributed-execution manager can be
    used without a `with` statement as follows::

        workflow = nvt.Workflow(["col"] >> ops.Normalize())
        dataset = nvt.Dataset(...)
        exec = Distributed()
        workflow.transform(dataset).to_parquet(...)
        exec.deactivate()

    Note that `deactivate()` must be used to resume default
    execution in the case that `Distributed` is not used in
    a `with` context.

    Since the default local cluster may be inefficient for
    many workflows, the user can also specify the specific
    `cluster_type` and `**cluster_options`. For example::

        with Distributed(
            cluster_type="cuda",
            force_new=True,  # Ignore existing cluster(s)
            n_workers=4,
            local_directory="/raid/dask-space",
            protocol="ucx",
            device_memory_limit=0.8,
            rmm_pool_size="20GB",
            log_spilling=True,
        ):
            workflow.transform(dataset).to_parquet(...)

    In this case, the `cluster_type="cuda"` calls for the
    creation of a `LocalCUDACluster`, and all other key-word
    arguments are passed to the `LocalCUDACluster` constructor.
    """

    def __init__(self, client=None, cluster_type=None, force_new=False, **cluster_options):
        self._initial_client = global_dask_client()  # Initial state
        self._client = client or "auto"  # Cannot be `None`
        self.cluster_type = cluster_type or ("cpu" if cuda is None else "cuda")
        self.cluster_options = cluster_options
        # We can only shut down the cluster in `shutdown`/`__exit__`
        # if we are generating it internally
        set_dask_client(self._client)
        self._allow_shutdown = global_dask_client() is None or force_new
        self._active = False
        self.force_new = force_new
        # Activate/deploy the client/cluster
        self._activate()

    @property
    def client(self):
        return self._client

    @property
    def cluster(self):
        return self.client.cluster

    @property
    def dashboard_link(self):
        return self.client.dashboard_link

    def _activate(self):
        if not self._active:
            self._client = set_dask_client(
                self._client,
                new_cluster=self.cluster_type,
                force_new=self.force_new,
                **self.cluster_options,
            )
        self._active = True
        if self._client in ("auto", None):
            raise RuntimeError(f"Failed to deploy a new local {self.cluster_type} cluster.")

    def _deactivate(self):
        self._client = set_dask_client(self._initial_client)
        self._active = False

    def deactivate(self):
        if self._allow_shutdown and self._active:
            self._client.close()
        self._deactivate()

    def __enter__(self):
        self._activate()
        return self

    def __exit__(self, *args):
        self.deactivate()

    def __del__(self, *args):
        self.deactivate()


class Serial:
    """Serial-Execution Context Manager

    Examples
    --------
    The easiest way to use `Serial` is within a
    conventional `with` statement::

        from merlin.core.utils import Serial

        workflow = nvt.Workflow(["col"] >> ops.Normalize())
        dataset = nvt.Dataset(...)
        with Serial():
            workflow.transform(dataset).to_parquet(...)

    In this case, all Dask-based scheduling and execution
    required within the `with Serial()` block will be
    performed using the "synchronous" (single-threaded)
    scheduler.

    Alternatively, the serial-execution manager can be
    used without a `with` statement as follows::

        workflow = nvt.Workflow(["col"] >> ops.Normalize())
        dataset = nvt.Dataset(...)
        exec = Serial()
        workflow.transform(dataset).to_parquet(...)
        exec.deactivate()

    Note that `deactivate()` must be used to resume
    default execution in the case that `Serial` is
    not used in a `with` context.
    """

    def __init__(self):
        # Save the initial client setting and
        # activate serial-execution mode
        self._initial_client = global_dask_client()
        self._client = self._initial_client
        self._active = False
        self._activate()

    @property
    def client(self):
        return self._client

    def _activate(self):
        # Activate serial-execution mode.
        # This just means we are setting the
        # global dask client to `None`
        if not self._active:
            set_dask_client(None)
        self._active = True
        if global_dask_client() is not None:
            raise RuntimeError("Failed to activate serial-execution mode.")

    def deactivate(self):
        # Deactivate serial-execution mode.
        # This just means we are setting the
        # global dask client to the original setting.
        set_dask_client(self._initial_client)
        self._active = False
        if self._initial_client is not None and global_dask_client() is None:
            raise RuntimeError("Failed to deactivate serial-execution mode.")

    def __enter__(self):
        self._activate()
        return self

    def __exit__(self, *args):
        self.deactivate()


def set_client_deprecated(client, caller_str):
    warnings.warn(
        f"The `client` argument is deprecated from {caller_str} "
        f"and will be removed in a future version of NVTabular. By "
        f"default, a global client in the same python context will be "
        f"detected automatically, and `merlin.utils.set_dask_client` "
        f"(as well as `Distributed` and `Serial`) can be used for "
        f"explicit control.",
        FutureWarning,
    )
    set_dask_client(client)


def set_dask_client(client="auto", new_cluster=None, force_new=False, **cluster_options):
    """Set the Dask-Distributed client
    Parameters
    -----------
    client : {"auto", None} or `dask.distributed.Client`
        The client to use for distributed-Dask execution.
        If `"auto"` (default) the current python context will
        be searched for an existing client object. Specify
        `None` to disable distributed execution altogether.
    new_cluster : {"cuda", "cpu", None}
        Type of local cluster to generate in the case that
        `client="auto"` and a global dask client is not
        detected in the current python context. The "cuda"
        option corresponds to `dask_cuda.LocalCUDACluster`,
        while "cpu" corresponds to `distributed.LocalCluster`.
        Default is `None` (no local cluster is generated).
    force_new : bool
        Whether to force the creation of a new local cluster
        in the case that a global client object is already
        detected. Default is False.
    **cluster_options :
        Key-word arguments to pass to the local-cluster
        constructor specified by `new_cluster` (e.g.
        `n_workers=2`).
    """
    _merlin_dask_client.set(client)

    # Check if we need to deploy a new cluster
    if new_cluster and client is not None:
        base, cluster = {
            "cuda": ("dask_cuda", "LocalCUDACluster"),
            "cpu": ("distributed", "LocalCluster"),
        }.get(new_cluster, (None, None))
        if global_dask_client() is not None and not force_new:
            # Don't deploy a new cluster if one already exists
            warnings.warn(
                f"Existing Dask-client object detected in the "
                f"current context. New {new_cluster} cluster "
                f"will not be deployed. Set force_new to True "
                f"to ignore running clusters."
            )
        elif base and cluster:
            try:
                base = importlib.import_module(base)
            except ImportError as err:
                # ImportError should only occur for LocalCUDACluster,
                # but I'm making this general to be "safe"
                raise ImportError(
                    f"new_cluster={new_cluster} requires {base}. "
                    f"Please make sure this library is installed. "
                ) from err
            _merlin_dask_client.set(Client(getattr(base, cluster)(**cluster_options)))
        else:
            # Something other than "cuda" or "cpu" was specified
            raise ValueError(f"{new_cluster} not a supported option for new_cluster.")

    # Return the active client object
    active = _merlin_dask_client.get()
    return None if active == "auto" else active


def global_dask_client():
    # First, check _merlin_dask_client
    merlin_client = _merlin_dask_client.get()
    if merlin_client and merlin_client != "auto":
        if merlin_client.cluster and merlin_client.cluster.workers:
            # Active Dask client already known
            return merlin_client
        else:
            # Our cached client is no-longer
            # active, reset to "auto"
            merlin_client = "auto"
    if merlin_client == "auto":
        try:
            # Check for a global Dask client
            set_dask_client(get_client())
            return _merlin_dask_client.get()
        except ValueError:
            pass
    # Catch-all
    return None


def run_on_worker(func, *args, **kwargs):
    # Run a function on a Dask worker using `delayed`
    # execution (if a Dask client is detected)
    if global_dask_client():
        # There is a specified or global Dask client. Use it
        return dask.delayed(func)(*args, **kwargs).compute()
    # No Dask client - Use simple function call
    return func(*args, **kwargs)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
try:
    import cudf
    import dask_cudf
except ImportError:
    cudf = None
from dask.dataframe.core import new_dd_object
from dask.highlevelgraph import HighLevelGraph

from .dataset_engine import DatasetEngine


class DataFrameDatasetEngine(DatasetEngine):
    """DataFrameDatasetEngine allows NVT to interact with a dask_cudf.DataFrame object
    in the same way as a dataset on disk.
    """

    def __init__(self, ddf, cpu=False, moved_collection=None):
        # we're not calling the constructor of the base class - since it has a bunch of
        # of parameters that assumes we're using files. TODO: refactor
        # pylint: disable=super-init-not-called
        self._ddf = ddf
        self.cpu = cpu
        self.moved_collection = moved_collection or False

    def to_ddf(self, columns=None, cpu=None):
        # Check if we are using cpu
        cpu = self.cpu if cpu is None else cpu

        # Move data from gpu to cpu if necessary
        _ddf = self._move_ddf("cpu") if (cpu and not self.cpu) else self._ddf

        if isinstance(columns, list):
            return _ddf[columns]
        elif isinstance(columns, str):
            return _ddf[[columns]]
        return _ddf

    def to_cpu(self):
        if self.cpu:
            return
        self._ddf = self._move_ddf("cpu")
        self.cpu = True
        self.moved_collection = not self.moved_collection

    def to_gpu(self):
        if not self.cpu:
            return
        self._ddf = self._move_ddf("gpu")
        self.cpu = False
        self.moved_collection = not self.moved_collection

    @property
    def num_rows(self):
        return len(self._ddf)

    def _move_ddf(self, destination):
        """Move the collection between cpu and gpu memory."""
        _ddf = self._ddf
        if (
            self.moved_collection
            and isinstance(_ddf.dask, HighLevelGraph)
            and hasattr(_ddf.dask, "key_dependencies")
        ):
            # If our collection has already been moved, and if the
            # underlying graph is a `HighLevelGraph`, we can just
            # drop the last "from_pandas-..." layer if the current
            # destination is "cpu", or we can drop the last
            # "to_pandas-..." layer if the destination is "gpu".
            search_name = "from_pandas-" if destination == "cpu" else "to_pandas-"

            pandas_conversion_layer = None
            pandas_conversion_dep = None
            for k, v in _ddf.dask.dependents.items():
                if k.startswith(search_name) and v == set():
                    pandas_conversion_layer = k
                    break
            if pandas_conversion_layer:
                deps = list(_ddf.dask.dependencies[pandas_conversion_layer])
                if len(deps) == 1:
                    pandas_conversion_dep = deps[0]

            if pandas_conversion_layer and pandas_conversion_dep:
                # We have met the criteria to remove the last "from/to_pandas-" layer
                new_layers = {
                    k: v for k, v in _ddf.dask.layers.items() if k != pandas_conversion_layer
                }
                new_deps = {
                    k: v for k, v in _ddf.dask.dependencies.items() if k != pandas_conversion_layer
                }
                hlg = HighLevelGraph(
                    layers=new_layers,
                    dependencies=new_deps,
                    key_dependencies=_ddf.dask.key_dependencies,
                )

                _meta = (
                    _ddf._meta.to_pandas() if destination == "cpu" else cudf.from_pandas(_ddf._meta)
                )
                return new_dd_object(hlg, pandas_conversion_dep, _meta, _ddf.divisions)

        if destination == "cpu":
            # Just extend the existing graph to move the collection to cpu
            return _ddf.to_dask_dataframe()

        elif destination == "gpu":
            # Just extend the existing graph to move the collection to gpu
            return dask_cudf.from_dask_dataframe(_ddf)

        else:
            raise ValueError(f"destination {destination} not recognized.")



# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pathlib
from typing import Union

import fsspec
import numpy

from ..schema import ColumnSchema
from ..schema import Schema as MerlinSchema
from ..tags import Tags
from . import proto_utils, schema_bp
from .schema_bp import Feature, FeatureType, FixedShape, FloatDomain, IntDomain
from .schema_bp import Schema as ProtoSchema
from .schema_bp import ValueCount

DOMAIN_ATTRS = {FeatureType.INT: "int_domain", FeatureType.FLOAT: "float_domain"}
FEATURE_TYPES = {
    "int": FeatureType.INT,
    "uint": FeatureType.INT,
    "float": FeatureType.FLOAT,
}


class TensorflowMetadata:
    """
    Reads and writes Merlin schemas as `tensorflow-metadata` Protobuf and JSON files

    See https://github.com/tensorflow/metadata for details on this metadata serialization format
    """

    def __init__(self, schema: ProtoSchema = None):
        self.proto_schema = schema

    @classmethod
    def from_json(cls, json: Union[str, bytes]) -> "TensorflowMetadata":
        """Create a TensorflowMetadata schema object from a JSON string

        Parameters
        ----------
        json : Union[str, bytes]
            The JSON string to parse

        Returns
        -------
        TensorflowMetadata
            Schema object parsed from JSON

        """
        schema = ProtoSchema().from_json(json)
        return TensorflowMetadata(schema)

    @classmethod
    def from_json_file(cls, path: str) -> "TensorflowMetadata":
        """Create a TensorflowMetadata schema object from a JSON file

        Parameters
        ----------
        path : str
            Path to the JSON file to parse

        Returns
        -------
        TensorflowMetadata
            Schema object parsed from JSON file

        """
        return cls.from_json(_read_file(path))

    @classmethod
    def from_proto_text(cls, proto_text: str) -> "TensorflowMetadata":
        """Create a TensorflowMetadata schema object from a Protobuf text string

        Parameters
        ----------
        proto_text : str
            Protobuf text string to parse

        Returns
        -------
        TensorflowMetadata
            Schema object parsed from Protobuf text

        """
        from tensorflow_metadata.proto.v0 import schema_pb2

        schema = proto_utils.proto_text_to_better_proto(
            ProtoSchema(), proto_text, schema_pb2.Schema()
        )

        return TensorflowMetadata(schema)

    @classmethod
    def from_proto_text_file(cls, path: str, file_name="schema.pbtxt") -> "TensorflowMetadata":
        """Create a TensorflowMetadata schema object from a Protobuf text file

        Parameters
        ----------
        path : str
            Path to the directory containing the Protobuf text file to parse
        file_name : str
            Name of the schema file. Defaults to "schema.pbtxt".

        Returns
        -------
        TensorflowMetadata
            Schema object parsed from Protobuf text file

        """
        path = pathlib.Path(path) / file_name
        return cls.from_proto_text(_read_file(str(path)))

    def to_proto_text(self) -> str:
        """Convert this TensorflowMetadata schema object to a Proto text string

        Returns
        -------
        str
            Generated Proto text string

        """
        from tensorflow_metadata.proto.v0 import schema_pb2

        return proto_utils.better_proto_to_proto_text(self.proto_schema, schema_pb2.Schema())

    def to_proto_text_file(self, path: str, file_name="schema.pbtxt"):
        """Write this TensorflowMetadata schema object to a file as a Proto text string

        Parameters
        ----------
        path : str
            Path to the directory containing the Protobuf text file
        file_name : str
            Name of the output file. Defaults to "schema.pbtxt".
        path: str :

        """
        _write_file(self.to_proto_text(), path, file_name)

    def copy(self, **kwargs) -> "TensorflowMetadata":
        """Create a copy of this TensorflowMetadata schema object

        Returns
        -------
        TensorflowMetadata
            Copy of this TensorflowMetadata schema object

        """
        schema_copy = proto_utils.copy_better_proto_message(self.proto_schema, **kwargs)
        return TensorflowMetadata(schema_copy)

    @classmethod
    def from_merlin_schema(cls, schema: MerlinSchema) -> "TensorflowMetadata":
        """Convert a MerlinSchema object to a TensorflowMetadata schema object

        Parameters
        ----------
        schema : MerlinSchema
            Schema object to convert

        Returns
        -------
        TensorflowMetadata
            Schema converted to a TensorflowMetadata schema object

        """
        features = []
        for col_name, col_schema in schema.column_schemas.items():
            features.append(_pb_feature(col_schema))

        proto_schema = ProtoSchema(feature=features)

        return TensorflowMetadata(proto_schema)

    def to_merlin_schema(self) -> MerlinSchema:
        """Convert this TensorflowMetadata schema object to a MerlinSchema object

        Returns
        -------
        MerlinSchema
            Schema converted to MerlinSchema object

        """
        merlin_schema = MerlinSchema()

        for feature in self.proto_schema.feature:
            col_schema = _merlin_column(feature)
            merlin_schema.column_schemas[col_schema.name] = col_schema

        return merlin_schema

    def to_json(self) -> str:
        """Convert this TensorflowMetadata schema object to a JSON string

        Returns
        -------
        str
            Schema converted to a JSON string

        """
        return self.proto_schema.to_json()


def _pb_int_domain(column_schema):
    domain = column_schema.properties.get("domain")
    if domain is None:
        return None

    return IntDomain(
        name=domain.get("name", column_schema.name),
        min=domain.get("min", None),
        max=domain.get("max", None),
        is_categorical=(
            Tags.CATEGORICAL in column_schema.tags or Tags.CATEGORICAL.value in column_schema.tags
        ),
    )


def _pb_float_domain(column_schema):
    domain = column_schema.properties.get("domain")
    if domain is None:
        return None
    return FloatDomain(
        name=column_schema.name,
        min=domain.get("min", None),
        max=domain.get("max", None),
    )


def _dtype_name(column_schema):
    # TODO: Decide if we need this since we've standardized on numpy types
    if hasattr(column_schema.dtype, "kind"):
        return numpy.core._dtype._kind_name(column_schema.dtype)
    elif hasattr(column_schema.dtype, "item"):
        return type(column_schema.dtype(1).item()).__name__
    elif isinstance(column_schema.dtype, str):
        return column_schema.dtype
    elif hasattr(column_schema.dtype, "__name__"):
        return column_schema.dtype.__name__
    else:
        raise TypeError(f"unsupported dtype for column schema: {column_schema.dtype}")


def _pb_extra_metadata(column_schema):
    properties = {
        k: v for k, v in column_schema.properties.items() if k not in ("domain", "value_count")
    }
    properties["dtype_item_size"] = numpy.dtype(column_schema.dtype).itemsize * 8
    properties["is_list"] = column_schema.is_list
    properties["is_ragged"] = column_schema.is_ragged
    return schema_bp.Any().from_dict(properties)


def _pb_tag(column_schema):
    return [tag.value if hasattr(tag, "value") else tag for tag in column_schema.tags]


def _pb_feature(column_schema):
    feature = Feature(name=column_schema.name)

    feature = _set_feature_domain(feature, column_schema)

    if column_schema.is_list:
        value_count = column_schema.properties.get("value_count", {})
        min_length = value_count.get("min")
        max_length = value_count.get("max")

        if min_length and max_length and min_length == max_length:
            feature.shape = FixedShape(min_length)
        elif min_length and max_length and min_length < max_length:
            feature.value_count = ValueCount(min=min_length, max=max_length)
        else:
            feature.value_count = ValueCount(min=0, max=0)

    feature.annotation.tag = _pb_tag(column_schema)
    feature.annotation.extra_metadata.append(_pb_extra_metadata(column_schema))
    return feature


def _set_feature_domain(feature, column_schema):
    DOMAIN_CONSTRUCTORS = {
        FeatureType.INT: _pb_int_domain,
        FeatureType.FLOAT: _pb_float_domain,
    }

    pb_type = FEATURE_TYPES.get(_dtype_name(column_schema))
    if pb_type:
        feature.type = pb_type

        domain_attr = DOMAIN_ATTRS[pb_type]
        domain_fn = DOMAIN_CONSTRUCTORS[pb_type]
        domain = domain_fn(column_schema)
        if domain:
            setattr(feature, domain_attr, domain)

    return feature


def _merlin_domain(feature):
    domain = {}

    domain_attr = DOMAIN_ATTRS.get(feature.type)

    if domain_attr and proto_utils.has_field(feature, domain_attr):
        domain_value = getattr(feature, domain_attr)
        if hasattr(domain_value, "min") and hasattr(domain_value, "max"):
            domain["min"] = domain_value.min
            domain["max"] = domain_value.max

        if hasattr(domain_value, "is_categorical"):
            domain["is_categorical"] = domain_value.is_categorical

        if hasattr(domain_value, "name"):
            name = domain_value.name
            if name != feature.name:
                domain["name"] = domain_value.name

    return domain


def _merlin_value_count(feature):
    if proto_utils.has_field(feature, "value_count"):
        value_count = feature.value_count
        if value_count.min != value_count.max != 0:
            return {"min": value_count.min, "max": value_count.max}


def _merlin_properties(feature):
    extra_metadata = feature.annotation.extra_metadata
    if len(extra_metadata) > 1:
        raise ValueError(
            f"{feature.name}: extra_metadata should have 1 item, has \
            {len(feature.annotation.extra_metadata)}"
        )
    elif len(extra_metadata) == 1:
        properties = feature.annotation.extra_metadata[0].value

        if isinstance(properties, bytes):
            properties = schema_bp.Any(value=properties).to_dict()

    else:
        properties = {}

    domain = _merlin_domain(feature)
    if domain:
        properties["domain"] = domain

    value_count = _merlin_value_count(feature)
    if value_count:
        properties["value_count"] = value_count
        properties["is_list"] = True
        properties["is_ragged"] = value_count.get("min") != value_count.get("max")
    return properties


int_dtypes_map = {
    8: numpy.int8,
    16: numpy.int16,
    32: numpy.int32,
    64: numpy.int64,
}


float_dtypes_map = {
    16: numpy.float16,
    32: numpy.float32,
    64: numpy.float64,
}


def _merlin_dtype(feature, properties):
    dtype = None
    item_size = int(properties.get("dtype_item_size", 0)) or None
    if feature.type == FeatureType.INT:
        if item_size and item_size in int_dtypes_map:
            dtype = int_dtypes_map[item_size]
        else:
            dtype = numpy.int
    elif feature.type == FeatureType.FLOAT:
        if item_size and item_size in float_dtypes_map:
            dtype = float_dtypes_map[item_size]
        else:
            dtype = numpy.float
    return dtype


def _merlin_column(feature):
    name = feature.name
    tags = list(feature.annotation.tag) or []
    properties = _merlin_properties(feature)
    dtype = _merlin_dtype(feature, properties)

    is_list = properties.pop("is_list", False)
    is_ragged = properties.pop("is_ragged", False)
    properties.pop("dtype_item_size", False)

    domain = properties.get("domain")
    if domain and domain.pop("is_categorical", False):
        if Tags.CATEGORICAL not in tags:
            tags.append(Tags.CATEGORICAL)

    return ColumnSchema(name, tags, properties, dtype, is_list, is_ragged=is_ragged)


def _read_file(path: str):
    # TODO: Should we be using fsspec here too?
    path = pathlib.Path(path)
    if path.is_file():
        with open(path, "r") as f:
            contents = f.read()
    else:
        raise ValueError("Path is not file")

    return contents


def _write_file(contents: str, path: str, filename: str):
    fs = fsspec.get_fs_token_paths(path)[0]

    try:
        with fs.open(fs.sep.join([str(path), filename]), "w") as f:
            f.write(contents)
    except Exception as e:
        if not fs.isdir(path):
            raise ValueError(f"The path provided is not a valid directory: {path}") from e
        raise



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from uuid import uuid4

import numpy as np

from .writer import ThreadedWriter


class HugeCTRWriter(ThreadedWriter):
    def __init__(self, out_dir, suffix=".data", **kwargs):
        super().__init__(out_dir, **kwargs)
        self.suffix = suffix
        if self.use_guid:
            self.data_paths = [
                os.path.join(out_dir, f"{i}i.{uuid4().hex}{self.suffix}")
                for i in range(self.num_out_files)
            ]
        else:
            self.data_paths = [
                os.path.join(out_dir, f"{i}{self.suffix}") for i in range(self.num_out_files)
            ]
        self.data_writers = [open(f, "wb") for f in self.data_paths]
        # Reserve 64 bytes for header
        header = np.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=np.longlong)
        for i, writer in enumerate(self.data_writers):
            writer.write(header.tobytes())

    def _write_table(self, idx, data):
        # Prepare data format
        np_label = data[self.labels].to_pandas().astype(np.single).to_numpy()
        np_conts = data[self.conts].to_pandas().astype(np.single).to_numpy()
        nnz = np.intc(1)
        np_cats = data[self.cats].to_pandas().astype(np.uintc).to_numpy()
        # Write all the data samples
        for i, _ in enumerate(np_label):
            # Write Label
            self.data_writers[idx].write(np_label[i].tobytes())
            # Write conts (HugeCTR: dense)
            self.data_writers[idx].write(np_conts[i].tobytes())
            # Write cats (HugeCTR: Slots)
            for j, _ in enumerate(np_cats[i]):
                self.data_writers[idx].write(nnz.tobytes())
                self.data_writers[idx].write(np_cats[i][j].tobytes())

    def _write_thread(self):
        while True:
            item = self.queue.get()
            try:
                if item is self._eod:
                    break
                idx, data = item
                with self.write_locks[idx]:
                    self._write_table(idx, data)
            finally:
                self.queue.task_done()

    def _close_writers(self):
        for i, writer in enumerate(self.data_writers):
            if self.cats:
                # Write HugeCTR Metadata
                writer.seek(0)
                # error_check (0: no error check; 1: check_num)
                # num of samples in this file
                # Dimension of the labels
                # Dimension of the features
                # slot_num for each embedding
                # reserved for future use
                header = np.array(
                    [
                        0,
                        self.num_samples[i],
                        len(self.labels),
                        len(self.conts),
                        len(self.cats),
                        0,
                        0,
                        0,
                    ],
                    dtype=np.longlong,
                )
                writer.write(header.tobytes())
            writer.close()
        return None

    def _bytesio_to_disk(self):
        raise ValueError("hugectr binary format doesn't support PER_WORKER shuffle yet")



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Text, Union

import numpy as np
import pandas as pd

from .tags import Tags, TagSet


@dataclass(frozen=True)
class Domain:
    min: Union[int, float]
    max: Union[int, float]
    name: Optional[str] = None


@dataclass(frozen=True)
class ColumnSchema:
    """A schema containing metadata of a dataframe column."""

    name: Text
    tags: Optional[TagSet] = field(default_factory=TagSet)
    properties: Optional[Dict[str, any]] = field(default_factory=dict)
    dtype: Optional[object] = None
    is_list: bool = False
    is_ragged: bool = False

    def __post_init__(self):
        """Standardize tags and dtypes on initialization

        Raises:
            TypeError: If the provided dtype cannot be cast to a numpy dtype
        """
        tags = TagSet(self.tags)
        object.__setattr__(self, "tags", tags)

        try:
            if hasattr(self.dtype, "numpy_dtype"):
                dtype = np.dtype(self.dtype.numpy_dtype)
            elif hasattr(self.dtype, "_categories"):
                dtype = self.dtype._categories.dtype
            elif isinstance(self.dtype, pd.StringDtype):
                dtype = np.dtype("O")
            else:
                dtype = np.dtype(self.dtype)
        except TypeError as err:
            raise TypeError(
                f"Unsupported dtype {self.dtype}, unable to cast {self.dtype} to a numpy dtype."
            ) from err

        object.__setattr__(self, "dtype", dtype)

    def with_name(self, name: str) -> "ColumnSchema":
        """Create a copy of this ColumnSchema object with a different column name

        Parameters
        ----------
        name : str
            New column name

        Returns
        -------
        ColumnSchema
            Copied object with new column name

        """
        return ColumnSchema(
            name,
            tags=self.tags,
            properties=self.properties,
            dtype=self.dtype,
            is_list=self.is_list,
            is_ragged=self.is_ragged,
        )

    def with_tags(self, tags: Union[str, Tags]) -> "ColumnSchema":
        """Create a copy of this ColumnSchema object with different column tags

        Parameters
        ----------
        tags : Union[str, Tags]
            New column tags

        Returns
        -------
        ColumnSchema
            Copied object with new column tags

        """
        return ColumnSchema(
            self.name,
            tags=self.tags.override(tags),
            properties=self.properties,
            dtype=self.dtype,
            is_list=self.is_list,
            is_ragged=self.is_ragged,
        )

    def with_properties(self, properties: dict) -> "ColumnSchema":
        """Create a copy of this ColumnSchema object with different column properties

        Parameters
        ----------
        properties : dict
            New column properties

        Returns
        -------
        ColumnSchema
            Copied object with new column properties

        Raises
        ------
        TypeError
            If properties are not a dict

        """
        if not isinstance(properties, dict):
            raise TypeError("properties must be in dict format, key: value")

        # Using new dictionary to avoid passing old ref to new schema
        properties.update(self.properties)

        return ColumnSchema(
            self.name,
            tags=self.tags,
            properties=properties,
            dtype=self.dtype,
            is_list=self.is_list,
            is_ragged=self.is_ragged,
        )

    def with_dtype(self, dtype, is_list: bool = None, is_ragged: bool = None) -> "ColumnSchema":
        """Create a copy of this ColumnSchema object with different column dtype

        Parameters
        ----------
        dtype : np.dtype
            New column dtype
        is_list: bool :
            Whether rows in this column contain lists.
             (Default value = None)
        is_ragged: bool :
            Whether lists in this column have varying lengths.
             (Default value = None)

        Returns
        -------
        ColumnSchema
            Copied object with new column dtype

        """
        is_list = is_list if is_list is not None else self.is_list

        if is_list:
            is_ragged = is_ragged if is_ragged is not None else self.is_ragged
        else:
            is_ragged = False

        return ColumnSchema(
            self.name,
            tags=self.tags,
            properties=self.properties,
            dtype=dtype,
            is_list=is_list,
            is_ragged=is_ragged,
        )

    @property
    def int_domain(self) -> Optional[Domain]:
        return self._domain() if np.issubdtype(self.dtype, np.integer) else None

    @property
    def float_domain(self) -> Optional[Domain]:
        return self._domain() if np.issubdtype(self.dtype, np.floating) else None

    @property
    def value_count(self) -> Optional[Domain]:
        value_count = self.properties.get("value_count")
        return Domain(**value_count) if value_count else None

    def __merge__(self, other):
        col_schema = self.with_tags(other.tags)
        col_schema = col_schema.with_properties(other.properties)
        col_schema = col_schema.with_dtype(
            other.dtype, is_list=other.is_list, is_ragged=other.is_ragged
        )
        col_schema = col_schema.with_name(other.name)
        return col_schema

    def __str__(self) -> str:
        return self.name

    def _domain(self) -> Optional[Domain]:
        """ """
        domain = self.properties.get("domain")
        return Domain(**domain) if domain else None


class Schema:
    """A collection of column schemas for a dataset."""

    def __init__(self, column_schemas=None):
        column_schemas = column_schemas or {}

        if isinstance(column_schemas, dict):
            self.column_schemas = column_schemas
        elif isinstance(column_schemas, (list, tuple)):
            self.column_schemas = {}
            for column_schema in column_schemas:
                if isinstance(column_schema, str):
                    column_schema = ColumnSchema(column_schema)
                self.column_schemas[column_schema.name] = column_schema
        else:
            raise TypeError("The `column_schemas` parameter must be a list or dict.")

    @property
    def column_names(self):
        return list(self.column_schemas.keys())

    def apply(self, selector) -> "Schema":
        """Select matching columns from this Schema object using a ColumnSelector

        Parameters
        ----------
        selector : ColumnSelector
            Selector that describes which columns match

        Returns
        -------
        Schema
            New object containing only the ColumnSchemas of selected columns

        """
        if selector is not None:
            schema = Schema()
            if selector.names:
                schema += self.select_by_name(selector.names)
            if selector.tags:
                schema += self.select_by_tag(selector.tags)
            return schema
        return self

    def apply_inverse(self, selector) -> "Schema":
        """Select non-matching columns from this Schema object using a ColumnSelector

        Parameters
        ----------
        selector : ColumnSelector
            Selector that describes which columns match

        Returns
        -------
        Schema
            New object containing only the ColumnSchemas of selected columns

        """
        if selector:
            return self - self.select_by_name(selector.names)
        return self

    def select_by_tag(self, tags: Union[Union[str, Tags], List[Union[str, Tags]]]) -> "Schema":
        """Select matching columns from this Schema object using a list of tags

        Parameters
        ----------
        tags : List[Union[str, Tags]] :
            List of tags that describes which columns match

        Returns
        -------
        Schema
            New object containing only the ColumnSchemas of selected columns

        """
        if not isinstance(tags, (list, tuple)):
            tags = [tags]

        selected_schemas = {}

        for _, column_schema in self.column_schemas.items():
            if any(x in column_schema.tags for x in tags):
                selected_schemas[column_schema.name] = column_schema

        return Schema(selected_schemas)

    def remove_by_tag(self, tags: Union[Union[str, Tags], List[Union[str, Tags]]]) -> "Schema":
        if not isinstance(tags, (list, tuple)):
            tags = [tags]

        selected_schemas = {}

        for column_schema in self.column_schemas.values():
            if not any(x in column_schema.tags for x in tags):
                selected_schemas[column_schema.name] = column_schema

        return Schema(selected_schemas)

    def select_by_name(self, names: Union[List[str], str]) -> "Schema":
        """Select matching columns from this Schema object using a list of column names

        Parameters
        ----------
        names: List[str] :
            List of column names that describes which columns match

        Returns
        -------
        Schema
            New object containing only the ColumnSchemas of selected columns

        """
        if isinstance(names, str):
            names = [names]

        selected_schemas = {
            key: self.column_schemas[key] for key in names if self.column_schemas.get(key, None)
        }
        return Schema(selected_schemas)

    def remove_col(self, col_name: str) -> "Schema":
        """Remove a column from this Schema object by name

        Parameters
        ----------
        col_name : str
            Name of the column to remove

        Returns
        -------
        Schema
            This Schema object after the column is removed

        """
        if col_name in self.column_names:
            del self.column_schemas[col_name]
        return self

    def without(self, col_names: List[str]) -> "Schema":
        """Remove columns from this Schema object by name

        Parameters
        ----------
        col_names : List[str]
            Names of the column to remove

        Returns
        -------
        Schema
            New Schema object after the columns are removed

        """
        return Schema(
            [
                col_schema
                for col_name, col_schema in self.column_schemas.items()
                if col_name not in col_names
            ]
        )

    def get(self, col_name: str, default: ColumnSchema = None) -> ColumnSchema:
        """Get a ColumnSchema by name

        Parameters
        ----------
        col_name : str
            Name of the column to get
        default: ColumnSchema :
            Default value to return if column is not found.
             (Default value = None)

        Returns
        -------
        ColumnSchema
            Retrieved column schema (or default value, if not found)

        """
        return self.column_schemas.get(col_name, default)

    @property
    def first(self) -> ColumnSchema:
        """
        Returns the first ColumnSchema in the Schema. Useful for cases where you select down
        to a single column via select_by_name or select_by_tag, and just want the value

        Returns
        -------
        ColumnSchema
            The first column schema present in this Schema object

        Raises
        ------
        ValueError
            If this Schema object contains no column schemas
        """
        if not self.column_schemas:
            raise ValueError("There are no columns in this schema to call .first on")

        return next(iter(self.column_schemas.values()))

    def __getitem__(self, column_name):
        if isinstance(column_name, str):
            return self.column_schemas[column_name]
        elif isinstance(column_name, (list, tuple)):
            return Schema([self.column_schemas[col_name] for col_name in column_name])

    def __setitem__(self, column_name, column_schema):
        self.column_schemas[column_name] = column_schema

    def __iter__(self):
        return iter(self.column_schemas.values())

    def __len__(self):
        return len(self.column_schemas)

    def __repr__(self):
        return str([col_schema.__dict__ for col_schema in self.column_schemas.values()])

    def __eq__(self, other):
        if not isinstance(other, Schema) or len(self.column_schemas) != len(other.column_schemas):
            return False
        return self.column_schemas == other.column_schemas

    def __add__(self, other):
        if other is None:
            return self
        if not isinstance(other, Schema):
            raise TypeError(f"unsupported operand type(s) for +: 'Schema' and {type(other)}")

        col_schemas = []

        # must account for same columns in both schemas,
        # use the one with more information for each field
        keys_self_not_other = [
            col_name for col_name in self.column_names if col_name not in other.column_names
        ]

        for key in keys_self_not_other:
            col_schemas.append(self.column_schemas[key])

        for col_name, other_schema in other.column_schemas.items():
            if col_name in self.column_schemas:
                # check which one
                self_schema = self.column_schemas[col_name]
                col_schemas.append(self_schema.__merge__(other_schema))
            else:
                col_schemas.append(other_schema)

        return Schema(col_schemas)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if other is None:
            return self

        if not isinstance(other, Schema):
            raise TypeError(f"unsupported operand type(s) for -: 'Schema' and {type(other)}")

        result = Schema({**self.column_schemas})

        for key in other.column_schemas.keys():
            if key in self.column_schemas.keys():
                result.column_schemas.pop(key, None)

        return result



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import numpy as np
import pandas as pd
import pytest
from dask.dataframe import assert_eq
from dask.dataframe.io.demo import names as name_list

import merlin

cudf = pytest.importorskip("cudf")
dask_cudf = pytest.importorskip("dask_cudf")

# Require uavro and fastavro library.
# Note that fastavro is only required to write
# avro files for testing, while uavro is actually
# used by AvroDatasetEngine.
fa = pytest.importorskip("fastavro")
pytest.importorskip("uavro")


@pytest.mark.parametrize("part_size", [None, "1KB"])
@pytest.mark.parametrize("size", [100, 5000])
@pytest.mark.parametrize("nfiles", [1, 2])
def test_avro_basic(tmpdir, part_size, size, nfiles):
    # Define avro schema
    schema = fa.parse_schema(
        {
            "name": "avro.example.User",
            "type": "record",
            "fields": [
                {"name": "name", "type": "string"},
                {"name": "age", "type": "int"},
            ],
        }
    )

    # Write avro dataset with two files.
    # Collect block and record (row) count while writing.
    nblocks = 0
    nrecords = 0
    paths = [os.path.join(str(tmpdir), f"test.{i}.avro") for i in range(nfiles)]
    records = []
    for path in paths:
        names = np.random.choice(name_list, size)
        ages = np.random.randint(18, 100, size)
        data = [{"name": names[i], "age": ages[i]} for i in range(size)]
        with open(path, "wb") as f:
            fa.writer(f, schema, data)
        with open(path, "rb") as fo:
            avro_reader = fa.block_reader(fo)
            for block in avro_reader:
                nrecords += block.num_records
                nblocks += 1
                records += list(block)
    if nfiles == 1:
        paths = paths[0]

    # Read back with dask.dataframe
    df = merlin.io.Dataset(paths, part_size=part_size, engine="avro").to_ddf()

    # Check basic length and partition count
    if part_size == "1KB":
        assert df.npartitions == nblocks
    assert len(df) == nrecords

    # Full comparison
    expect = pd.DataFrame.from_records(records)
    expect["age"] = expect["age"].astype("int32")
    assert_eq(df.compute().reset_index(drop=True), expect)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import contextlib
import glob
import os
import random

import dask
import numpy as np
import pandas as pd

try:
    import cudf

    try:
        import cudf.testing._utils

        assert_eq = cudf.testing._utils.assert_eq
    except ImportError:
        import cudf.tests.utils

        assert_eq = cudf.tests.utils.assert_eq
except ImportError:
    cudf = None

    def assert_eq(a, b, *args, **kwargs):
        if isinstance(a, pd.DataFrame):
            return pd.testing.assert_frame_equal(a, b, *args, **kwargs)
        elif isinstance(a, pd.Series):
            return pd.testing.assert_series_equal(a, b, *args, **kwargs)
        else:
            return np.testing.assert_allclose(a, b)


import pytest
from dask.distributed import Client, LocalCluster

import merlin

allcols_csv = ["timestamp", "id", "label", "name-string", "x", "y", "z"]
mycols_csv = ["name-string", "id", "label", "x", "y"]
mycols_pq = ["name-cat", "name-string", "id", "label", "x", "y"]
mynames = [
    "Alice",
    "Bob",
    "Charlie",
    "Dan",
    "Edith",
    "Frank",
    "Gary",
    "Hannah",
    "Ingrid",
    "Jerry",
    "Kevin",
    "Laura",
    "Michael",
    "Norbert",
    "Oliver",
    "Patricia",
    "Quinn",
    "Ray",
    "Sarah",
    "Tim",
    "Ursula",
    "Victor",
    "Wendy",
    "Xavier",
    "Yvonne",
    "Zelda",
]

_CUDA_CLUSTER = None


@pytest.fixture(scope="module")
def client():
    cluster = LocalCluster(n_workers=2)
    client = Client(cluster)
    yield client
    client.close()
    cluster.close()


@contextlib.contextmanager
def get_cuda_cluster():
    from dask_cuda import LocalCUDACluster

    CUDA_VISIBLE_DEVICES = os.environ.get("CUDA_VISIBLE_DEVICES", "0")
    n_workers = min(2, len(CUDA_VISIBLE_DEVICES.split(",")))
    cluster = LocalCUDACluster(n_workers=n_workers)
    yield cluster
    cluster.close()


@pytest.fixture(scope="session")
def datasets(tmpdir_factory):
    _lib = cudf if cudf else pd
    _datalib = cudf if cudf else dask
    df = _datalib.datasets.timeseries(
        start="2000-01-01",
        end="2000-01-04",
        freq="60s",
        dtypes={
            "name-cat": str,
            "name-string": str,
            "id": int,
            "label": int,
            "x": float,
            "y": float,
            "z": float,
        },
    ).reset_index()

    if _datalib is dask:
        df = df.compute()

    df["name-string"] = _lib.Series(np.random.choice(mynames, df.shape[0])).astype("O")

    # Add two random null values to each column
    imax = len(df) - 1
    for col in df.columns:
        if col in ["name-cat", "label", "id"]:
            break
        for _ in range(2):
            rand_idx = random.randint(1, imax - 1)
            if rand_idx == df[col].shape[0] // 2:
                # dont want null in median
                rand_idx += 1
            df[col].iloc[rand_idx] = None

    datadir = tmpdir_factory.mktemp("data_test")
    datadir = {
        "parquet": tmpdir_factory.mktemp("parquet"),
        "csv": tmpdir_factory.mktemp("csv"),
        "csv-no-header": tmpdir_factory.mktemp("csv-no-header"),
        "cats": tmpdir_factory.mktemp("cats"),
    }

    half = int(len(df) // 2)

    # Write Parquet Dataset
    df.iloc[:half].to_parquet(str(datadir["parquet"].join("dataset-0.parquet")), chunk_size=1000)
    df.iloc[half:].to_parquet(str(datadir["parquet"].join("dataset-1.parquet")), chunk_size=1000)

    # Write CSV Dataset (Leave out categorical column)
    df.iloc[:half].drop(columns=["name-cat"]).to_csv(
        str(datadir["csv"].join("dataset-0.csv")), index=False
    )
    df.iloc[half:].drop(columns=["name-cat"]).to_csv(
        str(datadir["csv"].join("dataset-1.csv")), index=False
    )
    df.iloc[:half].drop(columns=["name-cat"]).to_csv(
        str(datadir["csv-no-header"].join("dataset-0.csv")), header=False, index=False
    )
    df.iloc[half:].drop(columns=["name-cat"]).to_csv(
        str(datadir["csv-no-header"].join("dataset-1.csv")), header=False, index=False
    )

    return datadir


@pytest.fixture(scope="function")
def paths(engine, datasets):
    return sorted(glob.glob(str(datasets[engine]) + "/*." + engine.split("-")[0]))


@pytest.fixture(scope="function")
def df(engine, paths):
    _lib = cudf if cudf else pd
    if engine == "parquet":
        df1 = _lib.read_parquet(paths[0])[mycols_pq]
        df2 = _lib.read_parquet(paths[1])[mycols_pq]
    elif engine == "csv-no-header":
        df1 = _lib.read_csv(paths[0], header=None, names=allcols_csv)[mycols_csv]
        df2 = _lib.read_csv(paths[1], header=None, names=allcols_csv)[mycols_csv]
    elif engine == "csv":
        df1 = _lib.read_csv(paths[0], header=0)[mycols_csv]
        df2 = _lib.read_csv(paths[1], header=0)[mycols_csv]
    else:
        raise ValueError("unknown engine:" + engine)

    gdf = _lib.concat([df1, df2], axis=0)
    gdf["id"] = gdf["id"].astype("int64")
    return gdf


@pytest.fixture(scope="function")
def dataset(request, paths, engine):
    try:
        gpu_memory_frac = request.getfixturevalue("gpu_memory_frac")
    except Exception:  # pylint: disable=broad-except
        gpu_memory_frac = 0.01

    try:
        cpu = request.getfixturevalue("cpu")
    except Exception:  # pylint: disable=broad-except
        cpu = False

    kwargs = {}
    if engine == "csv-no-header":
        kwargs["names"] = allcols_csv

    return merlin.io.Dataset(paths, part_mem_fraction=gpu_memory_frac, cpu=cpu, **kwargs)


def run_in_context(func, *args, context=None, **kwargs):
    # Convenience utility to execute a function within
    # a specific `context`.  For example, this can be
    # used to test that a function raises a `UserWarning`
    # by setting `context=pytest.warns(UserWarning)`
    if context is None:
        context = contextlib.suppress()
    with context:
        result = func(*args, **kwargs)
    return result


# Allow to pass devices as parameters
def pytest_addoption(parser):
    parser.addoption("--devices", action="store", default="0", help="0,1,..,n-1")


@pytest.fixture
def devices(request):
    return request.config.getoption("--devices")



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import math
import queue
import threading
from typing import Optional

try:
    import cupy as cp
except ImportError:
    cp = None

import numpy as np
from fsspec.core import get_fs_token_paths

from merlin.core.dispatch import annotate

from .shuffle import shuffle_df


class Writer:
    def add_data(self, df):
        raise NotImplementedError()

    def package_general_metadata(self):
        raise NotImplementedError()

    @classmethod
    def write_general_metadata(cls, data, fs, out_dir, schema):
        raise NotImplementedError()

    @classmethod
    def write_special_metadata(cls, data, fs, out_dir):
        raise NotImplementedError()

    def close(self):
        pass


class ThreadedWriter(Writer):
    def __init__(
        self,
        out_dir,
        num_out_files=30,
        num_threads=0,
        cats=None,
        conts=None,
        labels=None,
        shuffle=None,
        fs=None,
        use_guid=False,
        bytes_io=False,
        cpu=False,
        fns=None,
        suffix=None,
    ):
        # set variables
        self.out_dir = out_dir
        self.cats = cats
        self.conts = conts
        self.labels = labels
        self.shuffle = shuffle
        self.column_names = None
        if labels and conts:
            self.column_names = labels + conts

        self.col_idx = {}

        self.num_threads = num_threads
        self.fns = [fns] if isinstance(fns, str) else fns
        if self.fns:
            # If specific file names were specified,
            # ignore `num_out_files` argument
            self.num_out_files = len(self.fns)
        else:
            self.num_out_files = num_out_files
        self.num_samples = [0] * self.num_out_files

        self.data_paths = None
        self.need_cal_col_names = True
        self.use_guid = use_guid
        self.bytes_io = bytes_io
        self.cpu = cpu
        self.suffix = suffix

        # Resolve file system
        self.fs = fs or get_fs_token_paths(str(out_dir))[0]

        # Only use threading if num_threads > 1
        self.queue = None
        if self.num_threads > 1:
            # create thread queue and locks
            self.queue = queue.Queue(num_threads)
            self.write_locks = [threading.Lock() for _ in range(num_out_files)]

            # signifies that end-of-data and that the thread should shut down
            self._eod = object()

            # create and start threads
            for _ in range(num_threads):
                write_thread = threading.Thread(target=self._write_thread, daemon=True)
                write_thread.start()

    def set_col_names(self, labels, cats, conts):
        self.cats = cats
        self.conts = conts
        self.labels = labels
        self.column_names = labels + conts

    def _write_table(self, idx, data):
        return

    def _write_thread(self):
        return

    @annotate("add_data", color="orange", domain="merlin_python")
    def add_data(self, df):

        # Early return
        if isinstance(df, list) and not df:
            return

        # Populate columns idxs
        if not self.col_idx:
            _df = df[0] if isinstance(df, list) else df
            for i, x in enumerate(_df.columns.values):
                self.col_idx[str(x)] = i

        if isinstance(df, list):
            # Data is already partitioned into a list
            self._add_data_list(df)
        elif self.num_out_files == 1:
            # Only writing to a single file. No need to
            # scatter or slice the data before writing
            self._add_single_file(df)
        else:
            # Use different mechanism to decompose and write each df
            # partition, depending on the backend (pandas or cudf).
            if self.cpu:
                self._add_data_slice(df)
            else:
                self._add_data_scatter(df)

        # wait for all writes to finish before exiting
        # (so that we aren't using memory)
        if self.num_threads > 1:
            self.queue.join()

    def _add_data_scatter(self, gdf):
        """Write scattered pieces.

        This method is for cudf-backed data only.
        """
        assert not self.cpu

        # Generate `ind` array to map each row to an output file.
        # This approach is certainly more optimized for shuffling
        # than it is for non-shuffling, but using a single code
        # path is probably worth the (possible) minor overhead.
        nrows = gdf.shape[0]
        typ = np.min_scalar_type(nrows * 2)
        if self.shuffle:
            ind = cp.random.choice(cp.arange(self.num_out_files, dtype=typ), nrows)
        else:
            ind = cp.arange(nrows, dtype=typ)
            cp.floor_divide(ind, math.ceil(nrows / self.num_out_files), out=ind)
        for x, group in enumerate(
            gdf.scatter_by_map(ind, map_size=self.num_out_files, keep_index=False)
        ):
            self.num_samples[x] += len(group)
            if self.shuffle:
                group = shuffle_df(group)
            if self.num_threads > 1:
                self.queue.put((x, group))
            else:
                self._write_table(x, group)

    def _add_data_slice(self, df):
        """Write shuffled slices.

        This method works for both pandas and cudf backends.
        """
        # Pandas does not support the `scatter_by_map` method
        # used in `_add_data_scatter`. So, we manually shuffle
        # the df and write out slices.
        if self.shuffle:
            df = shuffle_df(df)
        int_slice_size = df.shape[0] // self.num_out_files
        slice_size = int_slice_size if df.shape[0] % int_slice_size == 0 else int_slice_size + 1
        for x in range(self.num_out_files):
            start = x * slice_size
            end = start + slice_size
            # check if end is over length
            end = end if end <= df.shape[0] else df.shape[0]
            to_write = df.iloc[start:end]
            self.num_samples[x] = self.num_samples[x] + to_write.shape[0]
            if self.num_threads > 1:
                self.queue.put((x, to_write))
            else:
                self._write_table(x, to_write)

    def _add_data_list(self, dfs):
        """Write a list of DataFrames"""
        assert len(dfs) == self.num_out_files
        for x, df in enumerate(dfs):
            if self.shuffle:
                df = shuffle_df(df)
            self.num_samples[x] = self.num_samples[x] + df.shape[0]
            if self.num_threads > 1:
                self.queue.put((x, df))
            else:
                self._write_table(x, df)

    def _add_single_file(self, df):
        """Write to single file.

        This method works for both pandas and cudf backends.
        """
        df = shuffle_df(df) if self.shuffle else df
        self.num_samples[0] = self.num_samples[0] + df.shape[0]
        if self.num_threads > 1:
            self.queue.put((0, df))
        else:
            self._write_table(0, df)

    def package_general_metadata(self):
        data = {}
        if self.cats is None:
            return data
        data["data_paths"] = self.data_paths
        data["file_stats"] = []
        for i, path in enumerate(self.data_paths):
            fn = path.split(self.fs.sep)[-1]
            data["file_stats"].append({"file_name": fn, "num_rows": self.num_samples[i]})
        # cats
        data["cats"] = []
        for c in self.cats:
            data["cats"].append({"col_name": c, "index": self.col_idx[c]})
        # conts
        data["conts"] = []
        for c in self.conts:
            data["conts"].append({"col_name": c, "index": self.col_idx[c]})
        # labels
        data["labels"] = []
        for c in self.labels:
            data["labels"].append({"col_name": c, "index": self.col_idx[c]})

        return data

    @classmethod
    def write_general_metadata(cls, data, fs, out_dir, schema=None):
        if not data:
            return
        data_paths = data.pop("data_paths", [])
        num_out_files = len(data_paths)

        # Write file_list
        file_list_writer = fs.open(fs.sep.join([out_dir, "_file_list.txt"]), "w")
        file_list_writer.write(str(num_out_files) + "\n")
        for f in data_paths:
            file_list_writer.write(f + "\n")
        file_list_writer.close()

        # Write metadata json
        metadata_writer = fs.open(fs.sep.join([out_dir, "_metadata.json"]), "w")
        json.dump(data, metadata_writer)
        metadata_writer.close()

        # Write keyset file
        if schema:
            fs = get_fs_token_paths(out_dir)[0]
            with fs.open(fs.sep.join([out_dir, "_hugectr.keyset"]), "wb") as writer:
                for col in schema:
                    try:
                        for v in range(col.properties["embedding_sizes"]["cardinality"] + 1):
                            writer.write(v.to_bytes(4, "big"))
                    except KeyError:
                        pass

    @classmethod
    def write_special_metadata(cls, data, fs, out_dir):
        pass

    def _close_writers(self) -> Optional[dict]:
        for writer in self.data_writers:
            writer.close()
        return None

    def close(self):
        if self.num_threads > 1:
            # wake up all the worker threads and signal for them to exit
            for _ in range(self.num_threads):
                self.queue.put(self._eod)

            # wait for pending writes to finish
            self.queue.join()

        # Close writers and collect various metadata
        _general_meta = self.package_general_metadata()
        _special_meta = self._close_writers()  # pylint: disable=assignment-from-none

        # Move in-memory file to disk
        if self.bytes_io:
            _special_meta = self._bytesio_to_disk()

        return _general_meta, _special_meta

    def _bytesio_to_disk(self):
        raise NotImplementedError("In-memory buffering/shuffling not implemented for this format.")



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dask.utils import natural_sort_key
from fsspec.core import get_fs_token_paths


class DatasetEngine:
    """Base class for Dask-powered IO engines. Engines must provide a ``to_ddf`` method."""

    def __init__(self, paths, part_size, cpu=False, storage_options=None):
        paths = sorted(paths, key=natural_sort_key)
        self.paths = paths
        self.part_size = part_size
        self.storage_options = storage_options
        fs, fs_token, paths2 = get_fs_token_paths(
            paths, mode="rb", storage_options=self.storage_options
        )
        self.stripped_paths = paths2
        self.fs = fs
        self.fs_token = fs_token
        self.cpu = cpu

    def to_ddf(self, columns=None, cpu=None):
        raise NotImplementedError(""" Return a dask.dataframe.DataFrame or dask_cudf.DataFrame""")

    def to_cpu(self):
        raise NotImplementedError(""" Move data to CPU memory """)

    def to_gpu(self):
        raise NotImplementedError(""" Move data to GPU memory """)

    @property
    def _path_partition_map(self):
        return None

    @property
    def num_rows(self):
        raise NotImplementedError(""" Returns the number of rows in the dataset """)

    def validate_dataset(self, **kwargs):
        raise NotImplementedError(""" Returns True if the raw data is efficient for NVTabular """)

    @classmethod
    def regenerate_dataset(cls, dataset, output_path, columns=None, **kwargs):
        raise NotImplementedError(""" Regenerate a dataset with optimal properties """)

    def sample_data(self, n=1):
        """Return a sample of real data from the dataset

        Sample the partitions of the underlying Dask collection
        until a non-empty partition is found. Then, use the first
        ``n`` rows of that partition to infer dtype info. If no
        non-empty partitions are found, use the Dask metadata.
        """
        _ddf = self.to_ddf()
        for partition_index in range(_ddf.npartitions):
            _head = _ddf.partitions[partition_index].head(n)
            if len(_head):
                return _head
        return _ddf._meta



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum
from typing import List, Union


class Tags(Enum):
    """Standard tags used in the Merlin ecosystem"""

    # Feature types
    CATEGORICAL = "categorical"
    CONTINUOUS = "continuous"
    LIST = "list"
    SEQUENCE = "sequence"
    TEXT = "text"
    TEXT_TOKENIZED = "text_tokenized"
    TIME = "time"

    # Feature context
    USER = "user"
    USER_ID = "user_id"
    ITEM = "item"
    ITEM_ID = "item_id"
    SESSION = "session"
    SESSION_ID = "session_id"
    CONTEXT = "context"

    # Target related
    TARGET = "target"
    BINARY_CLASSIFICATION = "binary_classification"
    MULTI_CLASS_CLASSIFICATION = "multi_class"
    REGRESSION = "regression"


TAG_COLLISIONS = {
    Tags.CATEGORICAL: [Tags.CONTINUOUS],
    Tags.CONTINUOUS: [Tags.CATEGORICAL],
}


class TagSet:
    """Collection that normalizes tags and prevents collisions between incompatible tags"""

    def __init__(self, tags: List[Union[str, Tags]] = None):
        tags = tags or []
        if isinstance(tags, TagSet):
            tags = tags._tags

        self._tags = self._normalize_tags(tags)

        collisions = self._detect_collisions(self._tags, self._tags)
        if collisions:
            raise ValueError(
                f"Could not create a TagSet with the tags {self._tags}. "
                f"The following tags are incompatible: {collisions}"
            )

    def override(self, tags: List[Union[str, Tags]]) -> "TagSet":
        """Add new tags to the collection, removing any existing tags that are incompatible

        Parameters
        ----------
        tags : List[Union[str, Tags]] :
            Tags to add (and remove incompatibilities with)

        Returns
        -------
        TagSet
            A new combined set of tags with incompatible tags removed

        """
        tags = self._convert_to_tagset(tags)
        to_remove = self._detect_collisions(self._tags, tags)
        return TagSet(self - to_remove + tags)

    def __iter__(self):
        for tag in self._tags:
            yield tag

    def __len__(self):
        return len(self._tags)

    def __add__(self, tags):
        tags = self._convert_to_tagset(tags)
        return TagSet(self._tags.union(tags._tags))

    def __sub__(self, tags):
        tags = self._convert_to_tagset(tags)
        return TagSet(self._tags - tags._tags)

    def __eq__(self, tags):
        return self._tags == tags._tags

    def _detect_collisions(self, tags_a, tags_b):
        collisions = []
        for tag in tags_b:
            conflicting = TAG_COLLISIONS.get(tag, [])
            for conflict in conflicting:
                if conflict in tags_a:
                    collisions.append(conflict)
        return set(collisions)

    def _convert_to_tagset(self, tags):
        if not isinstance(tags, (list, set, TagSet)):
            tags = [tags]
        if not isinstance(tags, TagSet):
            tags = TagSet(tags)

        return tags

    def _normalize_tags(self, tags):
        return set(Tags[tag.upper()] if tag in Tags._value2member_map_ else tag for tag in tags)

    def __repr__(self) -> str:
        return str(self._tags)


TagsType = Union[TagSet, List[str], List[Tags], List[Union[Tags, str]]]



# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import TypeVar

import betterproto
from betterproto import Message as BetterProtoMessage
from google.protobuf import json_format, text_format
from google.protobuf.message import Message as ProtoMessage

ProtoMessageType = TypeVar("ProtoMessageType", bound=BetterProtoMessage)


def has_field(message: ProtoMessageType, field_name: str) -> bool:
    """Check if a Protobuf message has a particular field

    Parameters
    ----------
    message : ProtoMessageType
        Protobuf message object
    field_name : str
        Name of the field to look for
    message: ProtoMessageType :

    Returns
    -------
    bool
        `True` if the named field exists on the message object

    """
    return betterproto.serialized_on_wire(getattr(message, field_name))


def copy_better_proto_message(better_proto_message: ProtoMessageType, **kwargs) -> ProtoMessageType:
    """Create a copy of a Protobuf message

    Parameters
    ----------
    better_proto_message : ProtoMessageType
        The message to copy

    Returns
    -------
    ProtoMessageType
        Copy of better_proto_message

    """
    output = better_proto_message.__class__().parse(bytes(better_proto_message))
    for key, val in kwargs.items():
        setattr(output, key, val)

    return output


def better_proto_to_proto_text(
    better_proto_message: BetterProtoMessage, message: ProtoMessage
) -> str:
    """Convert a BetterProto message object to Protobuf text format

    Parameters
    ----------
    better_proto_message : BetterProtoMessage
        The message to convert
    message : ProtoMessage
        A blank (raw) Protobuf message object to parse into

    Returns
    -------
    str
        Protobuf text representation of better_proto_message

    """
    message.ParseFromString(bytes(better_proto_message))

    return text_format.MessageToString(message)


def proto_text_to_better_proto(
    better_proto_message: ProtoMessageType, proto_text: str, message: ProtoMessage
) -> ProtoMessageType:
    """Convert a Protobuf text format message into a BetterProto message object

    Parameters
    ----------
    better_proto_message : ProtoMessageType
        A BetterProto message object of the desired type
    proto_text : str
        The Protobuf text format message to convert
    message : ProtoMessage
        A blank (raw) Protobuf message object to parse proto_text into

    Returns
    -------
    ProtoMessageType
        A BetterProto message object containing attributes parsed from proto_text

    """
    proto = text_format.Parse(proto_text, message)
    json_str = json_format.MessageToJson(proto)

    return better_proto_message.__class__().from_json(json_str)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
import pytest

from merlin.core.dispatch import HAS_GPU, is_list_dtype, list_val_dtype, make_df

if HAS_GPU:
    _CPU = [True, False]
else:
    _CPU = [True]


@pytest.mark.parametrize("cpu", _CPU)
def test_list_dtypes(tmpdir, cpu):
    df = make_df(device="cpu" if cpu else "gpu")
    df["vals"] = [
        [[0, 1, 2], [3, 4], [5]],
    ]

    assert is_list_dtype(df["vals"])
    assert list_val_dtype(df["vals"]) == np.dtype(np.int64)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from merlin.dag.base_operator import BaseOperator as Operator
from merlin.dag.graph import Graph
from merlin.dag.selector import ColumnSelector
from merlin.schema import Schema


@pytest.mark.parametrize("engine", ["parquet"])
def test_graph_validates_schemas(dataset, engine):
    ops = ["a", "b", "c"] >> Operator()
    graph = Graph(ops)

    with pytest.raises(ValueError) as exc_info:
        graph.construct_schema(dataset.schema)

    assert "Missing column" in str(exc_info.value)


@pytest.mark.parametrize("engine", ["parquet"])
def test_compute_selector_validates_schemas(dataset, engine):
    op = Operator()
    schema = Schema(["a", "b"])
    selector = ColumnSelector(["c"])

    with pytest.raises(ValueError) as exc_info:
        op.compute_selector(schema, selector, ColumnSelector(), ColumnSelector())

    assert "Missing column" in str(exc_info.value)


@pytest.mark.parametrize("engine", ["parquet"])
def test_compute_input_schema_validates_schemas(dataset, engine):
    op = Operator()
    schema = Schema(["a", "b"])
    selector = ColumnSelector(["c"])

    with pytest.raises(ValueError) as exc_info:
        op.compute_input_schema(schema, Schema(), Schema(), selector)

    assert "Missing column" in str(exc_info.value)

    with pytest.raises(ValueError) as exc_info:
        op.compute_input_schema(Schema(), schema, Schema(), selector)

    assert "Missing column" in str(exc_info.value)

    with pytest.raises(ValueError) as exc_info:
        op.compute_input_schema(Schema(), Schema(), schema, selector)

    assert "Missing column" in str(exc_info.value)


@pytest.mark.parametrize("engine", ["parquet"])
def test_compute_output_schema_validates_schemas(dataset, engine):
    op = Operator()
    schema = Schema(["a", "b"])
    selector = ColumnSelector(["c"])

    with pytest.raises(ValueError) as exc_info:
        op.compute_output_schema(schema, selector)

    assert "Missing column" in str(exc_info.value)
